/*
 *  psychom_siml_base : Simulation of psychophysical adaptive threshold estimation
 *  Copyright (c) 2002 by Vit Drga <vit@psychophysics.org>.
 *                  
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 * 
 *   One internet location where the GNU General Public License MAY be found is
 *   http://www.gnu.org/licenses/gpl.txt
 *   although that location may change in the future.
 */
/*
------------------------
psychom_siml_base.c

  Simulation of threshold estimation using adaptive tracking procedures
  psychom_siml_base is a meta-program that calls psychom_siml_guts

Detailed documentation exists in a separate file:

  00_psychom_siml_README.txt

------------------------
Terminology:
   
 Throughout this code:
  "pfn" is short for "psychometric function",
  "reversal" and "changepoint" are synonymous
  "param" generally refers to a runtime parameter passed to psychom_siml_guts via stdin,
  "coeff" refers to a parameter of a given psychometric function, 
  e.g. the slope and intercept of a linear pfn would be called "coeffs", not "params".

*/

#include    <stdio.h>
#include    <math.h>
#include    <string.h>
#include    <stdlib.h> 
#include    <time.h>
#include    <limits.h>

/*******************************************************************************************/

long int getseed (char* filename)
{
  FILE* fptr;
  long int old_seed = -1, new_seed = -1;
  int error_ct = 0;
  int file_exists = 0;
  char ch;
  char tempstr[50];
  
  fptr = fopen(filename, "r");
  if (fptr == NULL) {
    file_exists = 0;
    old_seed = 7777777;
    new_seed = 7777777;
  } else {
    file_exists = 1;
    fscanf (fptr, "Globally across runs:  start_seed	end_seed", tempstr);
    fscanf (fptr, "%ld\t%ld\n", &old_seed, &new_seed);
  }
  if ((old_seed < 0) || (new_seed < 0)) {
    error_ct++;
    fprintf (stderr, "\n\nIn getseed: seeds not loaded from file %s. Halting.\n\n", filename);
  } else if (file_exists == 1) {
    fclose(fptr);
  }
  if(error_ct > 0)
    exit(1);
  else
    return new_seed;
}

/*******************************************************************************************/

int run_psychom_siml_guts (char* params)
{
  char com[1000];
  long int seed;
  int error_check = 0;

  strcpy (com, params);
  printf ("Executing:\n  %s\n", com);

  error_check = system (com);

  if (error_check == 127) {
    fprintf (stderr, "\nError in psychom_siml_base: system() call failed due to execve() call to /bin/sh\n");
  } else if (error_check == -1) {
    fprintf (stderr, "\nError in psychom_siml_base: system() call failed for unknown reasons\n");
  } else if (error_check != 0) {
    fprintf (stderr, 
      "\nIn psychom_siml_base: system() call worked, but psychom_siml_guts failed internally with error %d\n",
      error_check);
  }
  return error_check;
}

/***************************** **************************************************************/

void append_long_int_to_str (long int n, char* locstr, char* sepstr)
{
  char tempstr[20];

  sprintf(tempstr, "%ld", n); 
  strcat(locstr, sepstr);
  strcat(locstr, tempstr);
}

/*******************************************************************************************/

void append_double_to_str (double x, char* locstr, char* sepstr, int dp)
{
  char tempstr[40];

  if (dp == 4)
    sprintf(tempstr, "%.4f", x); 
  if (dp == 2)
    sprintf(tempstr, "%.2f", x); 
  else if (dp == 0)
    sprintf(tempstr, "%.0f", x);
  else 
    sprintf(tempstr, "%s", "-555555.555"); 

  strcat(locstr, sepstr);
  strcat(locstr, tempstr);
}

/*******************************************************************************************/

void set_com ( int n_initial_changepoints, 
               int n_final_changepoints, 
               int n_repeats, 
               double starting_level, 
               double starting_step_size, 
               double final_step_size,            
               int expt_opt,
               int adaptive_opt, 
               int pfn_opt,
               double stdev_cutoff_value,
               int save_separately_flag,
               int display_opt, 
               long int seed,
               int coeff_set_idx,
               char* pfn_coefficient_filename,
               char* com)
{  
/*
Here, the order of call is important. It should mirror the order of 
parameters passed to psychom_siml_guts. 
*/
  char sepstr[4];
  strcpy (sepstr, " ");
  strcpy (com, "psychom_siml_guts ");
  append_long_int_to_str ( n_initial_changepoints, com, sepstr);
  append_long_int_to_str ( n_final_changepoints,   com, sepstr);
  append_long_int_to_str ( n_repeats,              com, sepstr);
  append_double_to_str   ( starting_level,         com, sepstr, 2);
  append_double_to_str   ( starting_step_size,     com, sepstr, 2);
  append_double_to_str   ( final_step_size,        com, sepstr, 2);
  append_long_int_to_str ( expt_opt,               com, sepstr);
  append_long_int_to_str ( adaptive_opt,           com, sepstr);
  append_long_int_to_str ( pfn_opt,                com, sepstr);
  append_double_to_str   ( stdev_cutoff_value,     com, sepstr, 2);
  append_long_int_to_str ( save_separately_flag,   com, sepstr);
  append_long_int_to_str ( display_opt,            com, sepstr);
  append_long_int_to_str ( seed,                   com, sepstr);
  append_long_int_to_str ( coeff_set_idx,          com, sepstr);
  strcat (com, " ");
  strcat (com, pfn_coefficient_filename);
  strcat (com, " ");

}

/*******************************************************************************************/

/*

The parameters sent to psychom_siml_guts are listed below 
(in order):
  n_initial_changepoints  n_final_changepoints  n_repeats
  starting_level  starting_step_size  final_step_size
  expt_opt  adaptive_opt  pfn_opt
  stdev_cutoff_value
  save_separately_flag  display_opt
  seed
  coeff_set_idx  pfn_coefficient_filename
(all on one line)

The variables below have the same name and are of the same type
as in psychom_siml_guts.
You set their values in main() and pass them to set_com(),
which converts the values into a string called com[].
The string results_filename is handled separately.
After set_com() returns to main(), results_filename
is assigned in main() and appended to com[].

The final thing is to call
run_psychom_siml_guts(com);
and that's all.

NB: You can run_psychom_siml_guts() repeatedly from within a loop,
    setting new parameter values during each iteration.

    This is done in the example below using two nested loops. 
    There, results_filename is changed within the outer loop, 
    so that the multiple calls within the inner loop are all 
    written to the same results file, but each iteration of the 
    outer loop goes to a different results file.
    
    Make up your own variations. Play around with it.
*/


main (int argc, char** argv)
{

  time_t t1,t2;
  char com[900];
  char filename[1000];
  char filename_prefix[500];
  char pfn_coefficient_filename[1000];
  char results_filename[1000];
  char c0[20];
  char c1[20];
  char c2[20];
  char c3[20];
  char c4[20];
  int i, j, duration;
  int call_ct;
  int error_check = 0;

  int n_initial_changepoints, n_final_changepoints, n_repeats, expt_opt, adaptive_opt, pfn_opt;
  double starting_level, starting_step_size, final_step_size, stdev_cutoff_value;
  int save_separately_flag, display_opt, coeff_set_idx;
  long int seed;
  
  char sepstr[4];
  char tempstr[1000];

  char seedstr[20];

  (void) time(&t1);

  strcpy (sepstr, " ");
  n_initial_changepoints =     4;
  n_final_changepoints   =    12;
  n_repeats              = 30000;
  starting_level         =   0.0;
  starting_step_size     =   4.0;
  final_step_size        =   2.0;
  expt_opt               =     2;
  adaptive_opt           =     2;
  pfn_opt                =     5;
  stdev_cutoff_value     =  6.00;
  save_separately_flag   =     1;
  display_opt            =     5;
  seed                   =   123;
  coeff_set_idx          =     1;

  strcpy(pfn_coefficient_filename, "coeffs/test_pfn_coeffs_pfn5.txt");
  strcpy(pfn_coefficient_filename, "coeffs/Expt2001b_VD_s4kHz_50msgap_Stimpy_pfn_coeffs.txt");

  strcpy(filename_prefix, "Expt2001b_VD250Hz_28msgap_16_initial_steps__");
  strcpy(filename_prefix, "t1_");
  strcpy(filename_prefix, "output/Expt2001b_VD_s4kHz_50msgap_Stimpy_");

  strcpy (sepstr, " ");
  n_initial_changepoints =     4;
  n_final_changepoints   =    12;
  n_repeats              = 30000;
  starting_level         =   0.0;
  starting_step_size     =   4.0;
  final_step_size        =   2.0;
  expt_opt               =     2;
  adaptive_opt           =     2;
  pfn_opt                =     1;
  stdev_cutoff_value     =  6.00;
  save_separately_flag   =     1;
  display_opt            =     5;
  seed                   =   123;
  coeff_set_idx          =     1;

  strcpy(pfn_coefficient_filename, "coeffs/testIY_coeffs1.txt");
  strcpy(filename_prefix, "testIY_");

  call_ct = 0;
  for (j=0; j<1; j++) {
    starting_step_size = 4 * (j + 1);
    for (i=0; i<26; i++) {
      call_ct++;
      printf("\n\n\npsychom_siml_base calling psychom_simlguts, call no. %d\n\n", call_ct);
      starting_level = i * 5;
	    seed = getseed("psychom_randseed");
      set_com ( n_initial_changepoints, n_final_changepoints, n_repeats,
                starting_level, starting_step_size, final_step_size,
                expt_opt, adaptive_opt, pfn_opt,
                stdev_cutoff_value,
                save_separately_flag, display_opt,
                seed, 
                coeff_set_idx, pfn_coefficient_filename, com);
      strcat (com, filename_prefix);
/*      append_double_to_str (starting_step_size, com, "", 0);
      strcat (com, "dB_starting_stepsize");*/

      error_check = run_psychom_siml_guts (com);
      if (error_check != 0)
        i = j = INT_MAX-2; 
        /* to break out of both loops.  */
    }
  }

/* loop-breaking curio: don't set i = j = INT_MAX because i++ in the for loop 
 would then wrap i around from INT_MAX to INT_MIN (e.g. +2147483647 to -2147483648)
 which leaves i<0 true, and hence you never break out of the loop.  */


  (void) time(&t2);
  duration = difftime(t2, t1);

  printf("\n\n\npsychom_siml_base:  time taken = %d seconds\n", duration);
  printf("All done\n");

  exit(0);
}

/*******************************************************************************************/




















