/*
 *  psychom_siml_guts:
 *    Simulation of psychophysical adaptive threshold estimation
 *  Copyright (c) 2002 by Vit Drga <vit@psychophysics.org>.
 *
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 *   One internet location where the GNU General Public License MAY be found is
 *   www.gnu.org/licenses/gpl.txt
 *   although that location could change in the future.
 */
/*
------------------------

Usage summary:
Enter
  psychom_siml_guts -h
at the command prompt to get a usage summary.

Detailed documentation exists in a separate file:

  00_psychom_siml_README.txt

------------------------
Terminology:

 Throughout this code:
  "pfn" is short for "psychometric function",
  "reversal" and "changepoint" are synonymous
  "param" or "parameter" generally refers to a runtime parameter passed
    to this program via stdin,
  "coeff" or "coefficient" refers to a parameter of a given psychometric function

------------------------
Words of Wisdom:

The better is the enemy of the good
(Voltaire, 1764: "le mieux est l'ennemi du bien.")

------------------------
*/

#include    <stdio.h>
#include    <math.h>
#include    <string.h>
#include    <stdlib.h> 
#include    <time.h>
#include    <limits.h>

#include <errno.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>


#define    PFN_OPT              1
#define    COEFF_SET_IDX        1
#define    STDEV_CUTOFF_VALUE   6.0
#define    SAVE_SEPARATELY_FLAG 0

#define    UNLIMITED_STDEV      999.99

#define    DISPLAY_OPT         3
/*
DISPLAY_OPT: 1 = trial-by-trial, per-run summary, collation stats
DISPLAY_OPT: 2 = per-run summary and collation stats
DISPLAY_OPT: 3 = collation stats only (screen-friendly)
DISPLAY_OPT: 4 = collation stats only (spreadsheet-friendly)
DISPLAY_OPT: 5 = 3 to stdout, 4 to resultsfile(s)
*/

#define    MAX_NUM_REPEATS    30000
#define    MAX_NUM_CONDS      1
#define    MAX_NUM_PFNS       40

#define    INITIAL_LEVEL      36.0
#define    INITIAL_STEP_SIZE  4.0
#define    FINAL_STEP_SIZE    2.0

#define    NUM_CHANGE_1   20    /* no. reversals at initial step-size, not used for threshold mean */
#define    NUM_CHANGE_2   12    /* no. reversals at final   step-size, used for threshold mean */

#define    FIRST_MAIN_REVERSAL_IDX   (NUM_CHANGE_1)
#define    NUM_REVERSALS_USED        (NUM_CHANGE_2) 
#define    MAX_NUM_REVERSALS         (NUM_CHANGE_1 + NUM_CHANGE_2)
#define    LAST_MAIN_REVERSAL_IDX    (MAX_NUM_REVERSALS - 1)
#define    REVERSAL_SEPARATOR_IDX    (MAX_NUM_REVERSALS + 0)
#define    REVERSAL_MIN_IDX          (MAX_NUM_REVERSALS + 1)
#define    REVERSAL_MAX_IDX          (MAX_NUM_REVERSALS + 2)
#define    REVERSAL_RANGE_IDX        (MAX_NUM_REVERSALS + 3)
#define    REVERSAL_MEAN_IDX         (MAX_NUM_REVERSALS + 4)
#define    REVERSAL_STDEV_IDX        (MAX_NUM_REVERSALS + 5)
#define    REVERSAL_GEOMMEAN_IDX     (MAX_NUM_REVERSALS + 6)
#define    REVERSAL_GEOMSTDEV_IDX    (MAX_NUM_REVERSALS + 7)
#define    REVERSAL_HIGHEST_IDX      (REVERSAL_GEOMSTDEV_IDX)
/* always set REVERSAL_HIGHEST_IDX to equal the last of the IDXs set above */

#define    N_PROG_PARAMS  16
/* change N_PROG_PARAMS if adding in more command line opts */

#define    MAX_NUM_OVERRUNS   10
#define    MAX_TRIALS_PER_RUN   (MAX_NUM_REVERSALS * 12)
/* MAX_TRIALS_PER_RUN is from a rule of thumb that seems to work */


#define  NUM_PFN5_LINEAR_SEGMENTS  20
/* Used for pfn5() and used initialize_pfn(). 
  Change here if necessary and change the code in pfn5().
  Just make sure they're all consistent. 
*/

/*
NB:  n_UP_ONE_DOWN (positive flag vals) means the pfn should be a generally DECreasing function.
NB:  n_DOWN_ONE_UP (negitive flag vals) means the pfn should be a generally INCreasing function.
*/
#define    ADAPTIVE_TEST              999
#define    ADAPTIVE_TWO_UP_ONE_DOWN     2
#define    ADAPTIVE_TWO_DOWN_ONE_UP   (-2)
#define    ADAPTIVE_THREE_UP_ONE_DOWN   3
#define    ADAPTIVE_THREE_DOWN_ONE_UP (-3)
#define    ADAPTIVE_FOUR_UP_ONE_DOWN    4
#define    ADAPTIVE_FOUR_DOWN_ONE_UP  (-4)

#define    EXPT_2IFC    2
#define    EXPT_3IFC    3
#define    EXPT_4IFC    4
#define    EXPT_5IFC    5

#define    TRUE   1
#define    FALSE  0
#define    CORRECT            1
#define    INCORRECT          0

typedef struct { 
               double level;
               short int right_wrong;
							 } trial_details_type;

typedef  trial_details_type *run_arr_type;
typedef  run_arr_type all_run_arr_type[MAX_NUM_CONDS][MAX_NUM_REPEATS];

/* for details on the following, see www.function-pointer.org */
typedef double (*pfn_ptr_type) (double);

typedef struct { 
               double level;
               int trial_idx;
               int direction;
							 } reversal_details_type;

typedef reversal_details_type *reversal_arr_type;
typedef reversal_arr_type all_reversal_arr_type[MAX_NUM_CONDS][MAX_NUM_REPEATS];

/*
It ain't pretty, but below are global vars can be used for setting coefficients
for the psychometric functions. Only a few get used, depending on the pfn chosen
via pfn opt. Ultimately it boils down to a memory-speed tradeoff leaning towards speed.
See also setup_coeffs_ptr_arr();
*/

double chance_performance, perfect_performance;

#define   N_PFN_COEFFS   60
#define   ZX         -80808.08
double c0 = ZX;
double c1 = ZX;
double c2 = ZX;
double c3 = ZX;
double c4 = ZX;
double c5 = ZX;
double c6 = ZX;
double c7 = ZX;
double c8 = ZX;
double c9 = ZX;
double c10 = ZX;
double c11 = ZX;
double c12 = ZX;
double c13 = ZX;
double c14 = ZX;
double c15 = ZX;
double c16 = ZX;
double c17 = ZX;
double c18 = ZX;
double c19 = ZX;
double c20 = ZX;
double c21 = ZX;
double c22 = ZX;
double k0 = ZX;
double k1 = ZX;
double k2 = ZX;
double k3 = ZX;
double k4 = ZX;
double k5 = ZX;
double k6 = ZX;
double k7 = ZX;
double k8 = ZX;
double k9 = ZX;
double k10 = ZX;
double k11 = ZX;
double k12 = ZX;
double k13 = ZX;
double k14 = ZX;
double k15 = ZX;
double k16 = ZX;
double k17 = ZX;
double k18 = ZX;
double k19 = ZX;
double k20 = ZX;
double k21 = ZX;
double k22 = ZX;
double m0 = ZX;
double m1 = ZX;
double m2 = ZX;
double m3 = ZX;
double m4 = ZX;
double m5 = ZX;
double m6 = ZX;
double m7 = ZX;
double m8 = ZX;
double m9 = ZX;
double m10 = ZX;
double m11 = ZX;
double m12 = ZX;
double m13 = ZX;
double m14 = ZX;
double m15 = ZX;
double m16 = ZX;
double m17 = ZX;
double m18 = ZX;
double m19 = ZX;
double m20 = ZX;
double m21 = ZX;
double m22 = ZX;

/*
double step_size_arr[REVERSAL_HIGHEST_IDX+1];
*/

typedef double* ptr_to_double;
ptr_to_double coeffs_ptr_arr[N_PFN_COEFFS];  /* array of pointers to double */


/*******************************************************************************************/

void initialize_all_run_array (all_run_arr_type all_run_arr, int max_num_conds, int max_num_repeats)

{
  int i, j;
  
	for (i=0;i<max_num_conds;i++)
	  for (j=0;j<max_num_repeats;j++)
			all_run_arr[i][j] = NULL; 
}

/*******************************************************************************************/

void free_all_run_array (all_run_arr_type all_run_arr, int max_num_conds, int max_num_repeats)
{
  int i, j;
  
	for (i=0;i<max_num_conds;i++)
	  for (j=0;j<max_num_repeats;j++)
			if (all_run_arr[i][j] != NULL)
			  free(all_run_arr[i][j]); 
/*  printf("In free_all_run_array(): all freed\n"); */
}

/*******************************************************************************************/

void initialize_all_reversal_array (all_reversal_arr_type all_reversal_arr, int max_num_conds, int max_num_repeats)

{
  int i, j;
  
	for (i=0;i<max_num_conds;i++)
	  for (j=0;j<max_num_repeats;j++)
			all_reversal_arr[i][j] = NULL; 
}

/*******************************************************************************************/

void free_all_reversal_array (all_reversal_arr_type all_reversal_arr, int max_num_conds, int max_num_repeats)
{
  int i, j;
  
	for (i=0;i<max_num_conds;i++)
	  for (j=0;j<max_num_repeats;j++)
			if (all_reversal_arr[i][j] != NULL)
			  free(all_reversal_arr[i][j]); 
/*  printf("In free_all_reversal_array(): all freed\n"); */
}

/*******************************************************************************************/

void clean_up (all_reversal_arr_type all_reversal_arr, all_run_arr_type all_run_arr, 
               int max_num_conds, int max_num_repeats,
               int fptr_a_isopen, int fptr_b_isopen, FILE* fptr_a, FILE* fptr_b)
{
  free_all_reversal_array (all_reversal_arr, max_num_conds, max_num_repeats);
  free_all_run_array (all_run_arr, max_num_conds, max_num_repeats);

  if (fptr_a_isopen == 1)
    fclose (fptr_a);
  if (fptr_b_isopen == 1)
    fclose (fptr_b);
}

/*******************************************************************************************/

void init_run_array (run_arr_type* curr_run_arr, int n) 
{
  *curr_run_arr = calloc(n, sizeof(trial_details_type));
}

/*******************************************************************************************/

void init_reversal_array (reversal_arr_type* curr_trialblock, int n) 
{
  *curr_trialblock = calloc(n, sizeof(reversal_details_type));
 (*curr_trialblock)[REVERSAL_SEPARATOR_IDX].level = -9999999999.99;
 
}

/*******************************************************************************************/

/* test for my own peace of mind that ftell() returns an offset relative to SEEK_SET 
and not something else. This is what it does do, but the man page doesn't explicitly say so.
test_ftell() assumes *fptr is open and readable */

void test_ftell (FILE* fptr, int a, int b, int c)

{
  long i, ft1;
  char ch;
  long curr_foffset;
  
  fseek(fptr, (long)a, SEEK_SET);
  ft1 = ftell (fptr);
  fscanf(fptr, "%c", &ch);
  fseek(fptr, -1, SEEK_CUR);
  printf("\nIn test_ftell:  initial ftell = %2ld,   a = %2d,   ch = %c\n", ft1, a, ch);
  for (i=0; i<b; i++) 
    fscanf(fptr, "%c", &ch);
  ft1 = ftell (fptr);
  fscanf(fptr, "%c", &ch);
  fseek(fptr, -1, SEEK_CUR);
  printf("In test_ftell:     next ftell = %2ld,   b = %2d,   ch = %c\n", ft1, b, ch);
  for (i=0; i<c; i++)
    fscanf(fptr, "%c", &ch);
  ft1 = ftell (fptr);
  fscanf(fptr, "%c", &ch);
  fseek(fptr, -1, SEEK_CUR);
  printf("In test_ftell:    final ftell = %2ld,   c = %2d,   ch = %c\n\n", ft1, c, ch);

}

/*******************************************************************************************/

int check_errors_in_main (int max_num_conds, int max_num_repeats, 
      int max_num_reversals, int num_reversals_used, 
      int first_main_reversal_idx, int last_main_reversal_idx, int reversal_highest_index, 
      int argc, int expt_opt, int display_opt, int save_separately_flag)

{
  int error_ct = 0;
  int recompile_ct = 0;

/*
  if (reversal_highest_index  > REVERSAL_HIGHEST_IDX    ) {error_ct++; fprintf ( stderr, "error 1\n");}
  if (max_num_conds           > MAX_NUM_CONDS           ) {error_ct++; fprintf ( stderr, "error 2\n");}
  if (max_num_repeats         > MAX_NUM_REPEATS         ) {error_ct++; fprintf ( stderr, "error 3\n");}
  if (num_reversals_used      > NUM_REVERSALS_USED      ) {error_ct++; fprintf ( stderr, "error 4\n");}
  if (max_num_reversals       > MAX_NUM_REVERSALS       ) {error_ct++; fprintf ( stderr, "error 5\n");}
  if (last_main_reversal_idx  > LAST_MAIN_REVERSAL_IDX  ) {error_ct++; fprintf ( stderr, "error 6\n");}
  if (last_main_reversal_idx != max_num_reversals - 1   ) {error_ct++; fprintf ( stderr, "error 7\n");}
  if (max_num_reversals      != first_main_reversal_idx + num_reversals_used) {error_ct++; fprintf ( stderr, "error 8\n");}
  if ((argc != N_PROG_PARAMS+1) && (display_opt == 5))      {error_ct++; 
    fprintf ( stderr, "error 9.  Only use display_opt = 5 if passing all parameters to this program.\n");}
  if (reversal_highest_index  < 1 ) {error_ct++; fprintf ( stderr, "error 11\n");}
  if (max_num_conds           < 1 ) {error_ct++; fprintf ( stderr, "error 12\n");}
  if (max_num_repeats         < 1 ) {error_ct++; fprintf ( stderr, "error 13\n");}
  if (num_reversals_used      < 1 ) {error_ct++; fprintf ( stderr, "error 14\n");}
  if (max_num_reversals       < 1 ) {error_ct++; fprintf ( stderr, "error 15\n");}
  if (last_main_reversal_idx  < 0 ) {error_ct++; fprintf ( stderr, "error 16\n");}
  if ((save_separately_flag  != 0) && (save_separately_flag  != 1)) {error_ct++; fprintf ( stderr, "error 17\n");}
  if ((display_opt < 0) || (5 < display_opt)) {error_ct++; fprintf ( stderr, "error 18\n");}
*/

  if (reversal_highest_index > REVERSAL_HIGHEST_IDX) {
    error_ct++; 
    recompile_ct++;
    fprintf(stderr, "error 1:  reversal_highest_index (%d) > REVERSAL_HIGHEST_IDX (%d)\n", 
            reversal_highest_index,  REVERSAL_HIGHEST_IDX);
  }
  if (max_num_conds > MAX_NUM_CONDS) {
    error_ct++; 
    recompile_ct++;
    fprintf(stderr, "error 2: max_num_conds (%d) > MAX_NUM_CONDS (%d)\n", 
            max_num_conds, MAX_NUM_CONDS);
  }
  if (max_num_repeats > MAX_NUM_REPEATS) {
    error_ct++; 
    recompile_ct++;
    fprintf(stderr, "error 3: max_num_repeats (%d) > MAX_NUM_REPEATS (%d)\n", 
            max_num_repeats, MAX_NUM_REPEATS);
  }
  if (num_reversals_used > NUM_REVERSALS_USED) {
    error_ct++; 
    recompile_ct++;
    fprintf(stderr, "error 4: num_reversals_used (%d) > NUM_REVERSALS_USED (%d)\n", 
            num_reversals_used, NUM_REVERSALS_USED);
  }
  if (max_num_reversals > MAX_NUM_REVERSALS) {
    error_ct++; 
    recompile_ct++;
    fprintf(stderr, "error 5: max_num_reversals (%d) > MAX_NUM_REVERSALS (%d)\n", 
            max_num_reversals, MAX_NUM_REVERSALS);
  }
  if (last_main_reversal_idx > LAST_MAIN_REVERSAL_IDX) {
    error_ct++; 
    recompile_ct++;
    fprintf(stderr, "error 6: last_main_reversal_idx (%d) > LAST_MAIN_REVERSAL_IDX (%d)\n", 
            last_main_reversal_idx, LAST_MAIN_REVERSAL_IDX);
  }
  if (last_main_reversal_idx != max_num_reversals-1) {
    error_ct++; 
    fprintf(stderr, "error 7: last_main_reversal_idx (%d) != max_num_reversals-1 (%d)\n", 
            last_main_reversal_idx, max_num_reversals-1);
  }
  if (max_num_reversals != first_main_reversal_idx + num_reversals_used) {
    error_ct++; 
    fprintf(stderr, "error 8: max_num_reversals (%d) != first_main_reversal_idx+num_reversals_used (%d)\n",
            max_num_reversals, first_main_reversal_idx+num_reversals_used);
  }
  if ((argc != N_PROG_PARAMS+1) && (display_opt == 5)) {
    error_ct++; 
    fprintf(stderr, "error 9.  Only use display_opt = 5 if passing results_filename to this program.\n");
  }
  if (reversal_highest_index < 1) {
    error_ct++; 
    fprintf(stderr, "error 10: reversal_highest_index (%d) > 1\n", reversal_highest_index);
  }
  if (max_num_conds < 1) {
    error_ct++; 
    fprintf(stderr, "error 11: max_num_conds (%d) < 1\n", max_num_conds);
  }
  if (max_num_repeats < 1) {
    error_ct++; 
    fprintf(stderr, "error 12: max_num_repeats (%d) < 1\n", max_num_repeats);
  }
  if (num_reversals_used < 1) {
    error_ct++; 
    fprintf(stderr, "error 13: num_reversals_used (%d) < 1\n", num_reversals_used);
  }
  if (max_num_reversals < 1) {
    error_ct++; 
    fprintf(stderr, "error 14: max_num_reversals (%d) < 1\n", max_num_reversals);
  }
  if (last_main_reversal_idx < 0) {
    error_ct++; 
    fprintf(stderr, "error 15: last_main_reversal_idx (%d) < 1\n", last_main_reversal_idx);
  }
  if ((save_separately_flag != 0) && (save_separately_flag != 1)) {
    error_ct++; 
    fprintf(stderr, "error 16: save_separately_flag (%d) is not 0 or 1\n", save_separately_flag);
  }
  if ((display_opt < 0) || (5 < display_opt)) {
    error_ct++; 
    fprintf(stderr, "error 17: display_opt (%d) is not in the range 1-5\n", display_opt);
  }
  if (expt_opt <= 1) {
    error_ct++; 
    fprintf(stderr, "error 18: expt_opt (%d) is out of range 1-5\n", display_opt);
  }

  if (recompile_ct > 0) {
    fprintf(stderr, "If you are seeing this message,\n");
    fprintf(stderr, "it means that one or more program variables (lowercase)\n");
    fprintf(stderr, "exceeded the compiled maximum constant value (capitals).\n");
    fprintf(stderr, "Either reduce the program variable value,\n");
    fprintf(stderr, "or increase the maximum constant in the source code,\n");
    fprintf(stderr, "then recompile and rerun with the same program arguements\n");
  }
  return error_ct;
}

/*******************************************************************************************/

# define vM_SQRT1_2  1.414213562

/* The constant M_SQRT1_2 available in the GNU/linux setup
is not immediately avalable in my cygwin setup,
so I'll just use vM_SQRT1_2 as defined above
*/

/*
Below, cumul_Z_norm () returns Phi(x) = standard_normal_cdf(x).

M_SQRT1_2 = 1.0/square_root_of_two 
and erf() come from math.h

Phi(x) = standard_normal_cdf(x) = (1 + erf(x/sqrt(2)))/2,   x >= 0
where
erf (x) = 2/sqrt(pi) * integral from 0 to x of exp(-t^2) dt
So
Phi(x) = 0.5*(1.0 + (erf(x*M_SQRT1_2)));
for x >= 0.

Since 
Phi(-x) = 1.0 - Phi(x)
then
Phi(-x) = 1.0 - (0.5 + 0.5*(erf(x*M_SQRT1_2)));
        = 0.5*(1.0 - erf(x*M_SQRT1_2));
for x < 0.

For a great(!!) Gaussian inverse c.d.f., see
home.online.no/~pjacklam/notes/invnorm/

*/

double cumul_Z_norm (double z)

{
  if (z < 0)
    return 0.5*(1.0 + erf(z*vM_SQRT1_2));
  else
    return 0.5*(1.0 - erf(z*vM_SQRT1_2));
}


/*******************************************************************************************/

/* used to set up pfn5()  */

void setup_coeffs_ptr_arr(void)
{
  int i = 0;
  coeffs_ptr_arr[i] = &c0; i++;
  coeffs_ptr_arr[i] = &c1; i++;
  coeffs_ptr_arr[i] = &c2; i++;
  coeffs_ptr_arr[i] = &c3; i++;
  coeffs_ptr_arr[i] = &c4; i++;
  coeffs_ptr_arr[i] = &c5; i++;
  coeffs_ptr_arr[i] = &c6; i++;
  coeffs_ptr_arr[i] = &c7; i++;
  coeffs_ptr_arr[i] = &c8; i++;
  coeffs_ptr_arr[i] = &c9; i++;
  coeffs_ptr_arr[i] = &c10; i++;
  coeffs_ptr_arr[i] = &c11; i++;
  coeffs_ptr_arr[i] = &c12; i++;
  coeffs_ptr_arr[i] = &c13; i++;
  coeffs_ptr_arr[i] = &c14; i++;
  coeffs_ptr_arr[i] = &c15; i++;
  coeffs_ptr_arr[i] = &c16; i++;
  coeffs_ptr_arr[i] = &c17; i++;
  coeffs_ptr_arr[i] = &c18; i++;
  coeffs_ptr_arr[i] = &c19; i++;
  coeffs_ptr_arr[i] = &k0; i++;
  coeffs_ptr_arr[i] = &k1; i++;
  coeffs_ptr_arr[i] = &k2; i++;
  coeffs_ptr_arr[i] = &k3; i++;
  coeffs_ptr_arr[i] = &k4; i++;
  coeffs_ptr_arr[i] = &k5; i++;
  coeffs_ptr_arr[i] = &k6; i++;
  coeffs_ptr_arr[i] = &k7; i++;
  coeffs_ptr_arr[i] = &k8; i++;
  coeffs_ptr_arr[i] = &k9; i++;
  coeffs_ptr_arr[i] = &k10; i++;
  coeffs_ptr_arr[i] = &k11; i++;
  coeffs_ptr_arr[i] = &k12; i++;
  coeffs_ptr_arr[i] = &k13; i++;
  coeffs_ptr_arr[i] = &k14; i++;
  coeffs_ptr_arr[i] = &k15; i++;
  coeffs_ptr_arr[i] = &k16; i++;
  coeffs_ptr_arr[i] = &k17; i++;
  coeffs_ptr_arr[i] = &k18; i++;
  coeffs_ptr_arr[i] = &k19; i++;
  coeffs_ptr_arr[i] = &m0; i++;
  coeffs_ptr_arr[i] = &m1; i++;
  coeffs_ptr_arr[i] = &m2; i++;
  coeffs_ptr_arr[i] = &m3; i++;
  coeffs_ptr_arr[i] = &m4; i++;
  coeffs_ptr_arr[i] = &m5; i++;
  coeffs_ptr_arr[i] = &m6; i++;
  coeffs_ptr_arr[i] = &m7; i++;
  coeffs_ptr_arr[i] = &m8; i++;
  coeffs_ptr_arr[i] = &m9; i++;
  coeffs_ptr_arr[i] = &m10; i++;
  coeffs_ptr_arr[i] = &m11; i++;
  coeffs_ptr_arr[i] = &m12; i++;
  coeffs_ptr_arr[i] = &m13; i++;
  coeffs_ptr_arr[i] = &m14; i++;
  coeffs_ptr_arr[i] = &m15; i++;
  coeffs_ptr_arr[i] = &m16; i++;
  coeffs_ptr_arr[i] = &m17; i++;
  coeffs_ptr_arr[i] = &m18; i++;
  coeffs_ptr_arr[i] = &m19; i++;
}

/*******************************************************************************************/
void disp_coeffs (char* prestr, double k, double m, double c, int n)
{
  if ((k != ZX) || (m != ZX) || (c != ZX)) {
    if (n < 10)
      printf("%s  k%d = %11.4f    m%d = %11.4f    c%d = %11.4f\n", prestr, n, k, n, m, n, c);
    else
      printf("%s k%d = %11.4f   m%d = %11.4f   c%d = %11.4f\n", prestr, n, k, n, m, n, c);
  }

}

/*******************************************************************************************/

/*
For testing purposes:
Comment out the early return statement to make this function usable.
It's handy when testing a new pfn for double-checking that the coefficients 
are loaded and correctly assigned to k, c, and m variables.
*/

void show_all_coeffs(void)
{
  return;
  
  printf("for double-checking, the coeffs assigned to k, m, or c values are:\n");
  disp_coeffs ("In show_all_coeffs:", k0, m0, c0, 0);
  disp_coeffs ("In show_all_coeffs:", k1, m1, c1, 1);
  disp_coeffs ("In show_all_coeffs:", k2, m2, c2, 2);
  disp_coeffs ("In show_all_coeffs:", k3, m3, c3, 3);
  disp_coeffs ("In show_all_coeffs:", k4, m4, c4, 4);
  disp_coeffs ("In show_all_coeffs:", k5, m5, c5, 5);
  disp_coeffs ("In show_all_coeffs:", k6, m6, c6, 6);
  disp_coeffs ("In show_all_coeffs:", k7, m7, c7, 7);
  disp_coeffs ("In show_all_coeffs:", k8, m8, c8, 8);
  disp_coeffs ("In show_all_coeffs:", k9, m9, c9, 9);
  disp_coeffs ("In show_all_coeffs:", k10, m10, c10, 10);
  disp_coeffs ("In show_all_coeffs:", k11, m11, c11, 11);
  disp_coeffs ("In show_all_coeffs:", k12, m12, c12, 12);
  disp_coeffs ("In show_all_coeffs:", k13, m13, c13, 13);
  disp_coeffs ("In show_all_coeffs:", k14, m14, c14, 14);
  disp_coeffs ("In show_all_coeffs:", k15, m15, c15, 15);
  disp_coeffs ("In show_all_coeffs:", k16, m16, c16, 16);
  disp_coeffs ("In show_all_coeffs:", k17, m17, c17, 17);
  disp_coeffs ("In show_all_coeffs:", k18, m18, c18, 18);
  disp_coeffs ("In show_all_coeffs:", k19, m19, c19, 19);
  disp_coeffs ("In show_all_coeffs:", k20, m20, c20, 20);
  disp_coeffs ("In show_all_coeffs:", k21, m21, c21, 21);
  disp_coeffs ("In show_all_coeffs:", k22, m22, c22, 22);
}

/*******************************************************************************************/
/*
  load pfn coefficients from file into pfn_coeffs[].
  All coefficients are loaded as doubles (%lf, not %f)
  The array entry pfn_coeffs[0] is reserved for the number of coefficients per coeff set.
*/

int load_pfn_coeffs(int pfn_opt, double pfn_coeffs[], char* filename, int coeff_set_idx)
{
  FILE* fptr;
  struct stat buf;
  int err_check, err_check2, i, j, n_rows, n_cols;
  double dum;
  char ch;

  err_check = stat(filename, &buf);
  if (err_check < 0) {
    if (errno == ENOENT) 
      fprintf(stderr, "Error in load_pfn_coeffs: Pfn-coefficient file does not exist.\n");
    return (-1);
  }

  fptr = fopen (filename, "r");
  if (fptr == NULL) {
    fprintf(stderr, "Error in load_pfn_coeffs: Can't open file, errno = %d\n\n", errno);
    return (-1);
  }

/* Under gcc v2.96, linux_i386, EOF is a 4-byte signed int with the value EOF = -1 */
  err_check = 99;
  i = 0;
  pfn_coeffs[0] = 0;
  while ((err_check != EOF) && (err_check > 0)) {
    err_check = fscanf(fptr, "%c", &ch); 
    if (err_check > 0) {
      if (ch == '#') {
        do {
          err_check = fscanf(fptr, "%c", &ch);
        } while ((err_check != EOF) && (ch != '\n'));
      } else {
        fseek (fptr, -1, SEEK_CUR);
        break;
      }
    }
  }

/* NB from the user's point of view, 
   rows, cols and coeff_set_idx are all one-based indices, not zero-based */

  err_check = fscanf(fptr, " %d %d\n", &n_rows, &n_cols);
  if (err_check != 2) {
    fprintf(stderr, "Error in load_pfn_coeffs: The  n_rows n_cols  header numbers were not found.\n\n");
    return (-1);
  }

  if ((coeff_set_idx < 1) || (n_rows < coeff_set_idx)) {
    fprintf(stderr, "Error in load_pfn_coeffs: coefficient-set index out of range.\n\n");
    return (-1);
  }

  err_check2 = 0;
  coeff_set_idx--;  /* to make the index zero-based */

/* read off the first coeff_set_idx-1 sets */
  for (i=0; i<coeff_set_idx; i++)
    for (j=0; j<n_cols; j++) {
      err_check = fscanf(fptr, " %lf", &dum); 
      err_check2 = err_check2 + err_check;
  }
  
  if (err_check2 != coeff_set_idx * n_cols) {
    fprintf(stderr, "Error in load_pfn_coeffs: Didn't find selected coefficient set in coefficient file.\n\n", errno);
    return (-1);
  }
    
  err_check2 = 0;
  for (j=0; j<n_cols; j++) {
    err_check = fscanf(fptr, " %lf", &dum); 
    pfn_coeffs[j+1] = dum;
    err_check2 = err_check2 + err_check;
  }

  if (err_check2 != n_cols) {
    fprintf(stderr, "Error in load_pfn_coeffs: Didn't load all coefficients correctly.\n\n", errno);
    return (-1);
  }
    
  pfn_coeffs[0] = n_cols;  /* no. of pfn coefficients per set */
  printf("\n");
  printf("no. of coefficients =  %.0f  (stored in pfn_coeffs[0])\n", pfn_coeffs[0]);
  for (j=0; j<n_cols+1; j++)
    printf("pfn_coeffs[%2d] = %14.6f\n", j, pfn_coeffs[j]);

  fclose (fptr);
  return (0);
}

/*******************************************************************************************/
/* linear pfn */
double pfn1 (double x)
{
  return (c1 * x + c0);
}

/*******************************************************************************************/
/* quadratic pfn */
double pfn2 (double x)
{
  return (c2 * x * x + c1 * x + c0);
}

/*******************************************************************************************/
/* cubic pfn */
double pfn3 (double x)
{
  return (c3 * x * x * x + c2 * x * x + c1 * x + c0);
}

/*******************************************************************************************/
/* 
  piecewise-linear pfn with up to 20 linear sections
  
  The k-values k0...k19 represent the upper bound of each section,
  m0...m19 are the slopes, and c0..c19 are the intercepts.
  k19 is a special case: it's not used in practice and defaults to +infinity.

  Example:
  if you only need *FOUR* portions, then:
    set k0, k1, and k2 appropriately (along with their m and c values), 
    set k3 = *very-high* value (e.g. 1.0E20)
      (so that you can be sure x will never be tested against k4..k19),
    set m0 = 0.0,   (flat line)
    set m3 = 0.0,   (flat line)
    set c0 = perfect_performance 
    and c3 = chance_performance (for *DECreasing* p-fns)
    OR
    set c0 = chance_performance 
    and c3 = perfect_performance (for *INcreasing* p-fns)
*/

/* below: coded for simplicity, not efficiency */
double pfn5 (double x)
{
double y; 

  if      (x <= k0) y = m0 * x + c0;
  else if (x <= k1) y = m1 * x + c1;
  else if (x <= k2) y = m2 * x + c2;
  else if (x <= k3) y = m3 * x + c3;
  else if (x <= k4) y = m4 * x + c4;
  else if (x <= k5) y = m5 * x + c5;
  else if (x <= k6) y = m6 * x + c6;
  else if (x <= k7) y = m7 * x + c7;
  else if (x <= k8) y = m8 * x + c8;
  else if (x <= k9) y = m9 * x + c9;
  else if (x <= k10) y = m10 * x + c10;
  else if (x <= k11) y = m11 * x + c11;
  else if (x <= k12) y = m12 * x + c12;
  else if (x <= k13) y = m13 * x + c13;
  else if (x <= k14) y = m14 * x + c14;
  else if (x <= k15) y = m15 * x + c15;
  else if (x <= k16) y = m16 * x + c16;
  else if (x <= k17) y = m17 * x + c17;
  else if (x <= k18) y = m18 * x + c18;
  else               y = m19 * x + c19;

  return (y);
}
/*******************************************************************************************/

/* 

The next six functions, pfn12()...pfn16() and pfn22() are of similar form.
They are (piecewise from left to right): 
  two linear sections, a polynomial section, followed by two linear sections

In pfn12(), the polynomial is of the  2nd degree (quadratic)
In pfn13(), the polynomial is of the  3rd degree (cubic)
In pfn14(), the polynomial is of the  4th degree (quartic)
In pfn15(), the polynomial is of the  5th degree (quintic)
In pfn16(), the polynomial is of the  6th degree (??)
In pfn22(), the polynomial is of the 12th degree (??)
    
The polynomial coefficients are c10,c11,...,c22 and k10, where 
the nth coefficient, c(10+n), covers the nth degree within the polynomial
and k10 is the range upper-bound for the polynomial section of the pfn.

In pfn12,...,pfn16, only the c(10+n) values that are needed are used.
In pfn22, all 13 of the c(10+n) values must be assigned. 
All coefficients that will be used should be set to some value.
Any low-degree terms not used in a high-degree polynomial, 
should have their coefficients set to 0.0.

The polynomial terms are only ever ADDED together in the code.
For subtractive terms, make sure their coefficients are negative.

The linear portions use coefficient triplets (k0,m0,c0) and (k1,m1,c1) below the 
quadratic section and (k2,m2,c2) and (k3,m3,c3) above it.
If you just want one line segment at each end instead of two, say,
then set (k1,m1,c1) = (k0,m0,c0)  and  (k3,m3,c3) = (k2,m2,c2).
but NOTE: k3 may be set to any value in the coefficient file. 
It's handy for consistency, but it's NOT used in practice and defaults to +infinity.

This doesn't have to hold, but I'm expecting that for
the first linear section (y = m0 * x + c0), m0 = 0.0 and c0 = perfect_performance;
the  last linear section (y = m3 * x + c3), m3 = 0.0 and c3 = chance_performance.
or vice versa w.r.t. c0 and c3.

Generally,  k0 <= k1 <= k10 <= k2  should hold, but that is not tested for.

The 2nd-6th degree polynomials are done separately for speed.

*/

/*******************************************************************************************/

/* 2nd degree polynomial with two linear sections at each end. See notes before pfn12() */

double pfn12 (double x)
{ 
  double x_sqrd, y; 

  if (x <= k0)
    y = m0 * x + c0;
  else if (x <= k1)
    y = m1 * x + c1;
  else if (x <= k10)
    y = (c12 * x * x) + (c11 * x) + c10;
  else if (x <= k2) 
    y = m2 * x + c2;
  else  
    y = m3 * x + c3;
  return (y);
}

/*******************************************************************************************/

/* 3rd degree polynomial with two linear sections at each end. See notes before pfn12() */

double pfn13 (double x)
{ 
  double x_sqrd, y; 

  if (x <= k0)
    y = m0 * x + c0;
  else if (x <= k1)
    y = m1 * x + c1;
  else if (x <= k10) {
    x_sqrd = x * x;
    y = (c13 * x_sqrd * x) + (c12 * x_sqrd) + (c11 * x) + c10;
  } else if (x <= k2) 
    y = m2 * x + c2;
  else  
    y = m3 * x + c3;
  return (y);
}

/*******************************************************************************************/

/* 4th degree polynomial with two linear sections at each end. See notes before pfn12() */

double pfn14 (double x)
{ 
  double x_sqrd, y; 

  if (x <= k0)
    y = m0 * x + c0;
  else if (x <= k1)
    y = m1 * x + c1;
  else if (x <= k10) {
    x_sqrd = x * x;
    y = (c14 * x_sqrd * x_sqrd) + (c13 * x_sqrd * x) + (c12 * x_sqrd) + (c11 * x) + c10;
  } else if (x <= k2) 
    y = m2 * x + c2;
  else  
    y = m3 * x + c3;
  return (y);
}

/*******************************************************************************************/

/* 5th degree polynomial with two linear sections at each end. See notes before pfn12() */

double pfn15 (double x)
{ 
  double x_sqrd, y; 

  if (x <= k0)
    y = m0 * x + c0;
  else if (x <= k1)
    y = m1 * x + c1;
  else if (x <= k10) {
    x_sqrd = x * x;
    y = (c15 * x_sqrd * x_sqrd * x)  + 
        (c14 * x_sqrd * x_sqrd)      + 
        (c13 * x_sqrd * x)           + 
        (c12 * x_sqrd)               + 
        (c11 * x)                    + 
         c10;
  } else if (x <= k2) 
    y = m2 * x + c2;
  else  
    y = m3 * x + c3;
  return (y);
}

/*******************************************************************************************/

/* 6th degree polynomial with two linear sections at each end. See notes before pfn12() */

double pfn16 (double x)
{ 
  double x_sqrd, x_cubed,  y; 

  if (x <= k0)
    y = m0 * x + c0;
  else if (x <= k1)
    y = m1 * x + c1;
  else if (x <= k10) {
    x_sqrd = x * x;
    x_cubed = x_sqrd * x;
    y = (c16 * x_cubed * x_cubed) + 
        (c15 * x_cubed * x_sqrd)  + 
        (c14 * x_sqrd * x_sqrd)   + 
        (c13 * x_sqrd * x)        + 
        (c12 * x_sqrd)            + 
        (c11 * x)                 + 
         c10;
  } else if (x <= k2) 
    y = m2 * x + c2;
  else  
    y = m3 * x + c3;
  return (y);
}

/*******************************************************************************************/

/* 12th degree polynomial with two linear sections at each end. See notes before pfn12() */

/* Simplicity here, not efficiency. The pow() function might beat the multiplications on speed.
   Set coefficient value to 0.0 for all polynomial terms that are not used.
*/

double pfn22 (double x)
{ 
  double x_sqrd, x_cubed, x_fourth, y; 

  if (x <= k0)
    y = m0 * x + c0;
  else if (x <= k1)
    y = m1 * x + c1;
  else if (x <= k10) {
    x_sqrd = x * x;
    x_cubed = x_sqrd * x;
    x_fourth = x_sqrd * x_sqrd;  

    y = (c22 * x_fourth * x_fourth * x_fourth) + 
        (c21 * x_fourth * x_fourth * x_cubed)  + 
        (c20 * x_fourth * x_fourth * x_sqrd)   + 
        (c19 * x_fourth * x_fourth * x)        + 
        (c18 * x_fourth * x_fourth)            + 
        (c17 * x_fourth * x_cubed)             + 
        (c16 * x_fourth * x_sqrd)              + 
        (c15 * x_fourth * x)                   + 
        (c14 * x_fourth)                       + 
        (c13 * x_cubed)                        + 
        (c12 * x_sqrd)                         + 
        (c11 * x)                              + 
         c10;
  } else if (x <= k2) 
    y = m2 * x + c2;
  else  
    y = m3 * x + c3;
  return (y);
}

/*******************************************************************************************/

/* Sigmoidal pfn using a Gaussian cumulative distribution function

Coefficients are:
  m0 = mean
  c0 = stdev
  k0 = direction (any negative value for a decreasing pfn, 
       any positive value for a increasing pfn,
  k1 = perfect_performance - chance_performance;

The m, c, and k values are set in initialize_pfn(), 
and perfect_performance and chance_performance are set in init_expt_opt().

pfn6() returns an asymptotic value for any z-scores beyond +4 or -4.
*/

#define PFN6_Z_CUTOFF  4.0

double pfn6 (double x)

{ 
  double z, y;
  
  z = (x - m0) / c0;
  if (k0 >= 0.0) {  /* increasing sigmoidal pfn */
    if (fabs(z) < PFN6_Z_CUTOFF) {
      if (z < 0.0)    
        y = k1 * cumul_Z_norm(z) + chance_performance;
      else 
        y = k1 * cumul_Z_norm(z) + chance_performance;
    } else {
      if (z < 0.0)    
        y = chance_performance;
      else 
        y = perfect_performance;
    }
  } else {       /* decreasing sigmoidal pfn */
    if (fabs(z) < PFN6_Z_CUTOFF) {
      if (z < 0.0)    
        y = k1 * (1.0 - cumul_Z_norm(z)) + chance_performance;
      else 
        y = k1 * (1.0 - cumul_Z_norm(z)) + chance_performance;
    } else {
      if (z < 0.0)    
        y = perfect_performance;
      else 
        y = chance_performance;
    }
  }

  return (y);
}

/*******************************************************************************************/

/* bouncing p-fn, piecewise two linear parts, a quadratic part, followed by a linear part*/
double old_pfn11 (double x)
{
double x_sqrd, y; 

k0 = -2.583;
m0 = 0.0;
c0 = perfect_performance;


k1 = 45.0;
m1 = -0.575;
c1 = 98.515;

k2 = 96.0;
m2 = 0.0;
c2 = chance_performance;

c10 = -415.1686631910;
c11 = 30.4841335977;
c12 = -0.7110902778;
c13 = 0.007387037;
c14 = -0.00002875;

  if (x <= k0)
    y = m0 * x + c0;
  else if (x <= k1)
    y = m1 * x + c1;
  else if (x <= k2) {
    x_sqrd = x * x;
    y = (c14 * x_sqrd * x_sqrd) + (c13 * x_sqrd * x) + (c12 * x_sqrd) + (c11 * x) + c10;
  } else 
    y = m2 * x + c2;

  return (y);
}

/*******************************************************************************************/
/*"IY_expt3_ES__suppressor_40dB_4.75kHz__onfreq_masker_4000Hz__msgap_70ms__pfn_opt3_"*/
double old_pfn3 (double x)
{
double y; 
  if (x <= 92.54) {
    y = -0.29088 * x + 110.528;
  } else if (x <= 100.90) {
    y = -4.02388 * x + 456.00;
  }    
  else y = 50.0;
  return (y);
}

/*******************************************************************************************/
/*"IY_expt3_ES__suppressor_40dB_4.75kHz__onfreq_masker_4000Hz__msgap_60ms__pfn_opt4_"*/
double old_pfn4 (double x)
{
double y; 
  if (x <= 85.439) {
    y = -0.5314 * x + 121.4163;
  } else if (x <= 102.9222) {
    y = -1.4881 * x + 203.1563;
  }    
  else y = 50.0;
  return (y);
}

/*******************************************************************************************/

/* 
To add a new pfn:
  pick a number N that hasn't been used up til now (see beginning of code below),
  add a new pfn_arr[N]=pfnN; assignment near the top,
  add a new  case N: section to the main switch(pfn_opt) statement which assigns 
    k, c, and m coefficients from the array pfn_coeffs[]  
    (don't forget to break; the switch),
  and, finally, write the code for the pfn as a separate function somewhere above here.

  Also, update documentation in the file 00_psychom_siml_README.txt.

  When testing new code,
    use show_all_coeffs() called in initialize_pfn() to double-check that loaded coefficients 
    are assigned correctly. (uncomment an early return in show_all_coeffs() to make it usable.)

  Below, perfect_performance and chance_performance are global variables, 
    set in value via init_expt_opt() and used in various pfns.
*/

pfn_ptr_type initialize_pfn (int pfn_opt, pfn_ptr_type pfn_arr[], double pfn_coeffs[], int disp_opt, char* pfn_name)

{
  pfn_ptr_type pfn_ptr = NULL;
  int i;
  int num_coeffs;
  int num_segments;
  int degree;
  int error_ct;
  char tempstr[8];

  for(i=0; i<MAX_NUM_PFNS; i++) 
    pfn_arr[i] = NULL;

	pfn_arr[1]=pfn1;
	pfn_arr[2]=pfn2;
	pfn_arr[3]=pfn3;
	pfn_arr[5]=pfn5;
	pfn_arr[6]=pfn6;
	pfn_arr[12]=pfn12;
	pfn_arr[13]=pfn13;
	pfn_arr[14]=pfn14;
	pfn_arr[15]=pfn15;
	pfn_arr[16]=pfn16;
	pfn_arr[22]=pfn22;

  if ((pfn_arr[pfn_opt] == NULL) || (pfn_opt > MAX_NUM_PFNS-1)) {
    fprintf(stderr, "Error in initialize_pfn;  pfn_opt = %d,  pfn_arr[%d] = NULL. \n\n", pfn_opt, pfn_opt);
    return NULL;
  }
  
  printf("\n");
  num_coeffs = (int)(pfn_coeffs[0]+1.0E-9);  /* This works. Don't trust plain (int)float conversions */
  switch (pfn_opt) {
    case 1:  /* linear */
      strcpy(pfn_name, "linear");
			pfn_ptr = pfn_arr[pfn_opt];
      c1 = pfn_coeffs[1];
      c0 = pfn_coeffs[2];
      if (disp_opt != 0) 
        printf("pfn[1] chosen: linear:  slope = %.4f,  intercept = %.4f\n", c1, c0);
			break;
    case 2:  /* quadratic */
      strcpy(pfn_name, "quadratic");
			pfn_ptr = pfn_arr[pfn_opt];
      c2 = pfn_coeffs[1];
      c1 = pfn_coeffs[2];
      c0 = pfn_coeffs[3];
      if (disp_opt != 0) 
        printf("pfn[2] chosen: quadratic:  c2 = %.4f,  c1 = %.4f,  c0 = %.4f\n", c2, c1, c0);
			break;
    case 3:  /* cubic */
      strcpy(pfn_name, "cubic");
			pfn_ptr = pfn_arr[pfn_opt];
      c3 = pfn_coeffs[1];
      c2 = pfn_coeffs[2];
      c1 = pfn_coeffs[3];
      c0 = pfn_coeffs[4];
      if (disp_opt != 0) 
        printf("pfn[3] chosen: cubic:  c3 = %.4f,   c2 = %.4f,  c1 = %.4f,  c0 = %.4f\n", c3,  c2, c1, c0);
			break;
    case 5: /* piecewise linear */
      strcpy(pfn_name, "piecewise-linear");
      setup_coeffs_ptr_arr();
      num_segments = num_coeffs/3;
      if ((num_coeffs < 1) || ((num_coeffs % 3) != 0) || (num_segments > NUM_PFN5_LINEAR_SEGMENTS)) {
        pfn_ptr = NULL;
        break;
      } else {
        for (i=0; i<num_segments; i++) {
          *(coeffs_ptr_arr[(1*NUM_PFN5_LINEAR_SEGMENTS)+(i*1)]) = pfn_coeffs[i*3+0+1];  /* k-variables 0, 1, 2...*/
          *(coeffs_ptr_arr[(2*NUM_PFN5_LINEAR_SEGMENTS)+(i*1)]) = pfn_coeffs[i*3+1+1];  /* m-variables 0, 1, 2...*/
          *(coeffs_ptr_arr[(0*NUM_PFN5_LINEAR_SEGMENTS)+(i*1)]) = pfn_coeffs[i*3+2+1];  /* c-variables 0, 1, 2...*/
        }
        if (disp_opt != 0) {
          printf("pfn[5] chosen: piecewise-linear,  num_segments = %2d,  total num_coeffs = %2d\n", num_segments, num_coeffs);
			    for (i=0; i<num_segments; i++) {
            printf("linear segment %2d:    k = %9.4f,    m = %9.4f,   c = %9.4f\n", 
                   i, pfn_coeffs[i*3+0+1], pfn_coeffs[i*3+1+1], pfn_coeffs[i*3+2+1]);
          }
          printf("\n");
        }
        pfn_ptr = pfn_arr[pfn_opt];
      }
			break;
    case 6:  /* Sigmoidal pfn using a Gaussian cumulative distribution function */
        strcpy(pfn_name, "Gaussian_cdf");
        m0 = pfn_coeffs[1];
        c0 = pfn_coeffs[2];
        k0 = pfn_coeffs[3];
        k1 = perfect_performance - chance_performance; 
/*       k1 is a scalar internal to this program, not a runtime program parameter.
         perfect_performance and chance_performance are set by init_expt_opt() 
         which is called in main() before initialize_pfn() is called in main().
*/
			break;
    case 12: /* falling through to 22... */
    case 13: /* falling through to 22... */
    case 14: /* falling through to 22... */
    case 15: /* falling through to 22... */
    case 16: /* falling through to 22... */
    case 22: /* polynomial (up to 5th degree) with piecewise linear sections appended*/
       degree = num_coeffs-14;
       error_ct = 0;
       if ( (degree < 2) || ((6 < degree) && (pfn_opt != 22)) )
          error_ct++;
       if ( (2 <= degree) && (degree <= 6) && (pfn_opt-10 != degree))
          error_ct++;
       if ( (2 <= pfn_opt-10) && (pfn_opt-10 <= 6) && (pfn_opt-10 != degree) )
          error_ct++;
       if (error_ct > 0) {
          fprintf(stderr, "Error in initialize_pfn for piecewise linear and polynomial mix:\n"); 
          fprintf(stderr, "  pfn_opt    = %2d (implied degree = %d),\n", pfn_opt, pfn_opt-10);
          fprintf(stderr, "  num_coeffs = %2d (implied degree = %d), \n", num_coeffs, num_coeffs-13);
          return NULL;
        } else {
          /* Four triplets for the linear sections, where k3 IS included in the parameter file.
             See the note before pfn12() for details.
          */
          strcpy(pfn_name, "linear_and_polynomial_degree_");
          sprintf(tempstr, "%d", degree);
          strcat(pfn_name, tempstr);
          k0 = pfn_coeffs[1];
          m0 = pfn_coeffs[2];
          c0 = pfn_coeffs[3];
          k1 = pfn_coeffs[4];
          m1 = pfn_coeffs[5];
          c1 = pfn_coeffs[6];
          k2 = pfn_coeffs[7];
          m2 = pfn_coeffs[8];
          c2 = pfn_coeffs[9];
          k3 = pfn_coeffs[10];
          m3 = pfn_coeffs[11];
          c3 = pfn_coeffs[12];

          k10 = pfn_coeffs[13];

          /* each c(10+n) below is the coefficient for the term of degree n
             Polynomial degrees above 2 are optional */
          c10 = pfn_coeffs[14];
          c11 = pfn_coeffs[15];
          c12 = pfn_coeffs[16];
          if (num_coeffs >= 17) c13 = pfn_coeffs[17];
          if (num_coeffs >= 18) c14 = pfn_coeffs[18];
          if (num_coeffs >= 19) c15 = pfn_coeffs[19];
          if (num_coeffs >= 20) c16 = pfn_coeffs[20];
          if (num_coeffs >= 21) c17 = pfn_coeffs[21];
          if (num_coeffs >= 22) c18 = pfn_coeffs[22];
          if (num_coeffs >= 23) c19 = pfn_coeffs[23];
          if (num_coeffs >= 24) c20 = pfn_coeffs[24];
          if (num_coeffs >= 25) c21 = pfn_coeffs[25];
          if (num_coeffs >= 26) c22 = pfn_coeffs[26];

          if (disp_opt != 0) {
            printf("pfn[%d] chosen: piecewise linear and polynomial of degree %d\n", pfn_opt, degree);
			      for (i=0; i<2; i++) {
              printf("  linear segment %d:    k = %9.4f,    m = %9.4f,   c = %9.4f\n", 
                     i, pfn_coeffs[i*3+0+1], pfn_coeffs[i*3+1+1], pfn_coeffs[i*3+2+1]);
            }
            printf("polynomial segment: ");
            for (i=0; i<degree+1; i++) {
              printf(" c%d = %.4f,",i, pfn_coeffs[13+i]);
            }
            printf("\n");
			      for (i=2; i<4; i++) {
              printf("  linear segment %d:    k = %9.4f,    m = %9.4f,   c = %9.4f\n", 
                     i, pfn_coeffs[i*3+0+1], pfn_coeffs[i*3+1+1], pfn_coeffs[i*3+2+1]);
            }
          }
          pfn_ptr = pfn_arr[pfn_opt];
        }
			break;
    default:  /*defaults to simple linear decreasing*/
      strcpy(pfn_name, "default");
			pfn_ptr = pfn_arr[0];  
      c1 = -2.0;
      c0 = 180.0;
			break;
  }

  show_all_coeffs();

  printf("\n");
  return pfn_ptr;
}

/*******************************************************************************************/
#define DISPLAYED_ARRAY_INDEX_BASE  1

void collate_reversal_arrs (all_reversal_arr_type all_reversal_arr, int max_num_conds, int max_num_repeats,
                            int first_main_reversal_idx, int last_main_reversal_idx, int num_reversals_used,
                            double stdev_cutoff_value,
                            double starting_level, double starting_step_size, double final_step_size,
                            long int start_seed, long int end_seed,
                            int expt_opt, int adaptive_opt, int pfn_opt, int display_opt,
                            char* expt_name, char* short_adaptive_name, char* pfn_name,
                            FILE *fptr1, FILE *fptr2,
                            char* pfn_coefficient_filename, int coeff_set_idx, double pfn_coeffs[])
{

/*
Used here so that the collation of index values (run length, first changepoint, etc.) 
are displayed AS IF the trial numbering followed 1,2,3... not 0,1,2...
The values are changed from zero--based to one-based at collation time, 
and not before during block-running time. DON'T change this because the 
zero-based indices indices are used earlier.
*/
  reversal_arr_type reversal_arr;
  int i, j, sweep;
  int n = 0;
  double sum = 0.0;
  double sos = 0.0;
  double soc = 0.0;
  double soq = 0.0;
  double min = +1.0E20;
  double max = -1.0E20;
  double sum_start  = 0.0;
  double sum_end    = 0.0;
  double sum_length = 0.0;
  double sos_start  = 0.0;
  double sos_end    = 0.0;
  double sos_length = 0.0;
  double rms_stdev = 0.0;
  double global_min = +1.0E20;
  double global_max = -1.0E20;
  int min_end = +99999999;
  int max_end = -99999999;
  double stderror;
/* Oh man... learning experience. I had used a variable called stderr,
 which the gcc compiler was happy with under linux, but not under mingw32.
 Now I know. Hence the variable is now called stderror.
*/
  double mean, stdev, temp, temp1, temp_stdev_cutoff;
  double mean_start, mean_end, mean_length, stdev_start, stdev_end, stdev_length;
  double percentage_of_blocks_converged;
  FILE *fptr;
  int n_pfn_coeffs;
  static int full_num_repeats;

  for (sweep=0; sweep<2; sweep++) {
    if ((sweep == 1) && (stdev_cutoff_value > UNLIMITED_STDEV))
      break;
      if (sweep == 0) {
        temp_stdev_cutoff = UNLIMITED_STDEV;
        fptr = fptr1;
      } else {
        temp_stdev_cutoff = stdev_cutoff_value;
        fptr = fptr2;
      }
          
    n = 0;
    sum = 0.0;
    sos = 0.0;
    soc = 0.0;
    soq = 0.0;
    min = 1.0E20;
    max = -1.0E20;
    sum_start  = 0.0;
    sum_end    = 0.0;
    sum_length = 0.0;
    sos_start  = 0.0;
    sos_end    = 0.0;
    sos_length = 0.0;
    rms_stdev = 0.0;
    global_min = +1.0E20;
    global_max = -1.0E20;
    min_end = +99999999;
    max_end = -99999999;
	  for (i=0;i<max_num_conds;i++) {
	    for (j=0;j<max_num_repeats;j++) {
        reversal_arr = all_reversal_arr[i][j];
        temp1 = reversal_arr[REVERSAL_STDEV_IDX].level;
        if(temp1 < temp_stdev_cutoff) {
          n++;
          temp = reversal_arr[REVERSAL_MEAN_IDX].level;
          sum = sum + temp;
          sos = sos + temp * temp;
          if (temp < global_min)
            global_min = temp;
          if (temp > global_max)
            global_max = temp;

          rms_stdev = rms_stdev + temp1 * temp1;

          temp = reversal_arr[first_main_reversal_idx].trial_idx + DISPLAYED_ARRAY_INDEX_BASE;
          sum_start = sum_start + temp;
          sos_start = sos_start + temp * temp;

          temp = reversal_arr[last_main_reversal_idx].trial_idx + DISPLAYED_ARRAY_INDEX_BASE;
          sum_end = sum_end + temp;
          sos_end = sos_end + temp * temp;
          if (temp < min_end)
            min_end = temp;
          if (temp > max_end)
            max_end = temp;

          temp = reversal_arr[REVERSAL_RANGE_IDX].trial_idx;
            /* It's okay. This array entry and field was used to store run-length values,
               NOT range values.. see note at the end of crunch_reversal_arr() */
          sum_length = sum_length + temp;
          sos_length = sos_length + temp * temp;
        }
      }
    }
    mean = sum / n;
    stdev = sqrt((sos - (n * mean * mean)) / (n - 1));
    stderror = stdev / sqrt(n);
    rms_stdev = sqrt(rms_stdev / n);

    mean_start = sum_start / n;
    stdev_start = sqrt((sos_start - (n * mean_start * mean_start)) / (n - 1));
    mean_end = sum_end / n;
    stdev_end = sqrt((sos_end - (n * mean_end * mean_end)) / (n - 1));
    mean_length = sum_length / n;
    stdev_length = sqrt((sos_length - (n * mean_length * mean_length)) / (n - 1));

    if (sweep == 0) 
      full_num_repeats = n;

    if ((display_opt == 1) || (display_opt == 2) || (display_opt == 3)) {
      fprintf (fptr, "---------------------------------------------------\n");
      fprintf (fptr, "global summary:\n");
      fprintf (fptr, "start_seed = %10ld,  end_seed = %10ld\n", start_seed, end_seed);
      fprintf (fptr, "expt_opt = %d,   adaptive_opt = %2d,   pfn_opt =%2d\n", expt_opt, adaptive_opt, pfn_opt);
      fprintf (fptr, "expt:   %s, adaptive rule: %s,   pfn: %s\n", expt_name, short_adaptive_name, pfn_name);
      fprintf (fptr, "non-counting changepoints: %3d per run\n", first_main_reversal_idx);
      fprintf (fptr, "        main changepoints: %3d per run\n\n", num_reversals_used);
      fprintf (fptr, "      stdev cutoff = %5.2f\n", temp_stdev_cutoff);
      fprintf (fptr, "    starting_level = %5.2f\n",starting_level);
      fprintf (fptr, "starting_step_size = %5.2f\n",starting_step_size);
      fprintf (fptr, "   final_step_size = %5.2f\n",final_step_size);

      fprintf (fptr, "\nn = %d runs\n", n);
      fprintf (fptr, "across runs:   mean = %5.4f,  stderr = %4.4f,  stdev = %4.4f,   min = %4.4f,  max = %4.4f\n", 
                     mean, stderror, stdev, global_min, global_max);
      fprintf (fptr, "\nwithin runs:  rms(stdev_per_run) = %4.4f\n\n", rms_stdev);
      fprintf (fptr, "first index:   mean = %6.2f,  stdev = %6.2f      (main changepoints start trial-no.)\n", mean_start, stdev_start);
      fprintf (fptr, "     length:   mean = %6.2f,  stdev = %6.2f      (run length for last %d changepoints)\n", 
                     mean_length, stdev_length, num_reversals_used);
      fprintf (fptr, "  end index:   mean = %6.2f,  stdev = %6.2f      (total run length, from first trial to last)\n", 
                     mean_end, stdev_end);
      fprintf (fptr, "  end index:    min = %6d,    max = %6d      (total run length, min and max)\n", min_end, max_end);
      fprintf (fptr, "(In collation above: index vals starting at one, not zero)\n");

      if (sweep == 1) {
        fprintf (fptr, "\nThere were %d runs out of %d (or %.2f%% of blocks) with stdevs less than %.2f\n",
                 n, full_num_repeats, 100.0*(double)(n)/full_num_repeats, temp_stdev_cutoff);
      }

      fprintf (fptr, "\n---------------------------------------------------\n");
    }

/*
Recommend leaving this code as is for ease (!) of reading.
If changing the output, then also update headerstr in main()
*/
    if (display_opt == 4) {
      n_pfn_coeffs = (int)(pfn_coeffs[0]+1.0E-9);
      percentage_of_blocks_converged = 100.0*(double)n/(double)max_num_repeats;
      fprintf (fptr, 
"%.2f\t%d\t%d\t%.2f\t\
%.4f\t%.4f\t%.4f\t\
%.4f\t%.4f\t%.4f\t\
%.4f\t%.4f\t%.4f\t\
%d\t%d\t\
%.2f\t%.2f\t%.2f\t%.2f\t%.2f\t%.2f\t\
%d\t%d\t\
%ld\t%ld\t\
%d\t%d\t%d\t\
%s\t%s\t%s\t\
%s\t%d\t%d",
temp_stdev_cutoff, n, max_num_repeats, percentage_of_blocks_converged, 
starting_level, starting_step_size, final_step_size, 
mean, stderror, stdev,
global_min, global_max, rms_stdev, 
first_main_reversal_idx, num_reversals_used, 
mean_start, stdev_start, mean_end, stdev_end, mean_length, stdev_length, 
min_end, max_end, 
start_seed, end_seed, 
expt_opt, adaptive_opt, pfn_opt, 
expt_name, short_adaptive_name, pfn_name,
pfn_coefficient_filename, coeff_set_idx, n_pfn_coeffs);

      for (i=0; i<n_pfn_coeffs; i++)
        fprintf (fptr, "\t%.8f",pfn_coeffs[i+1]);
      fprintf (fptr, "\n");
    }
  }

}

/*******************************************************************************************/

void crunch_reversal_arr (reversal_arr_type reversal_arr, int first_main_reversal_idx, int last_main_reversal_idx)
{
  int i;
  int n = 0;
  double sum = 0.0;
  double sos = 0.0;
  double soc = 0.0;
  double soq = 0.0;
  double min = 1.0E20;
  double max = -1.0E20;
  double stdev, mean, temp, temp2;
  
  for (i=first_main_reversal_idx; i<last_main_reversal_idx+1; i++) {
    n++;
    temp = reversal_arr[i].level;
    temp2 = temp * temp;
    sum = sum + temp;
    sos = sos + temp2;
    soc = soc + temp * temp2;
    soq = soq + temp2 * temp2;
    if (temp < min)
      min = temp;
    if (max < temp)
      max = temp;
  }
  mean = sum / n;
  stdev = sqrt((sos - (n * mean * mean)) / (n - 1));
  reversal_arr[REVERSAL_MIN_IDX].level = min;
  reversal_arr[REVERSAL_MAX_IDX].level = max;
  reversal_arr[REVERSAL_RANGE_IDX].level = max - min;
  reversal_arr[REVERSAL_MEAN_IDX ].level = mean;
  reversal_arr[REVERSAL_STDEV_IDX].level = stdev;

/* for these last five index vals, 
     the .level     field (used above) is used for min, max and range,
     the .trial_idx field (used below) is used for index storing, and NOT for min, max, etc...

  The values put in to the .trial_idx field are unrelated to the values in the .level field for the 
  same reversal_arr[] member structure.

Below: the values put in to the .trial_idx field hold the run_array indices of (respectively): 
  the first reversal after changing to final step size,
  the last reversal (= total run length index), 
  the no. of trials from the first changepoint to the last changepoint inclusive,
  the no. of initial changepoints (that won't count)
  the no. of final changepoints  (that will count)

NB: these indices are all zero-based. The adjustment to one-based trial numbers
only happens in collate_reversal_arrs().
*/
  reversal_arr[REVERSAL_MIN_IDX].trial_idx   = reversal_arr[first_main_reversal_idx].trial_idx;
  reversal_arr[REVERSAL_MAX_IDX].trial_idx   = reversal_arr[ last_main_reversal_idx].trial_idx;
  reversal_arr[REVERSAL_RANGE_IDX].trial_idx = reversal_arr[REVERSAL_MAX_IDX].trial_idx - 
                                               reversal_arr[REVERSAL_MIN_IDX].trial_idx + 1;
  reversal_arr[REVERSAL_MEAN_IDX].trial_idx  = first_main_reversal_idx;
  reversal_arr[REVERSAL_STDEV_IDX].trial_idx = last_main_reversal_idx - first_main_reversal_idx + 1;
}

/*******************************************************************************************/
void display_reversal_arr (reversal_arr_type reversal_arr, int reversal_highest_index, int max_num_reversals, 
                           int ii, int j, int tot_num_repeats, int display_opt, FILE *fptr, long int end_seed)
{
/* ii was intended as a condition_number parameter, but that's not used at present, so ignore it*/
  int i;
  if (display_opt == 1) {
    fprintf (fptr, "\nchangepoints:\n");
    fprintf (fptr, "  i   idx       X   direction\n");
    for (i=0; i<max_num_reversals; i++)
      fprintf (fptr, "%3d   %3d    %5.2f    %2d\n", i, reversal_arr[i].trial_idx, reversal_arr[i].level, reversal_arr[i].direction);
    for (; i<reversal_highest_index+1; i++)
      fprintf (fptr, "%3d   %3d    %5.2f    %2d\n", i, reversal_arr[i].trial_idx, reversal_arr[i].level, reversal_arr[i].direction);
  }
  if ((1 <= display_opt) && (display_opt <= 2)) {
    fprintf (fptr, "run %2d of %2d,  mean = %5.2f,  stdev = %4.2f,   min = %5.2f,  max = %5.2f,  end_seed = %10ld\n", 
      j+1, tot_num_repeats, reversal_arr[REVERSAL_MEAN_IDX].level, reversal_arr[REVERSAL_STDEV_IDX].level,
      reversal_arr[REVERSAL_MIN_IDX].level, reversal_arr[REVERSAL_MAX_IDX].level, end_seed);
  }
}

/*******************************************************************************************/

void init_expt_opt (int expt_opt, double* perfect_performance, double* chance_performance, char* expt_name)
{
  strcpy(expt_name,"");
  sprintf (expt_name, "%d", expt_opt);
  strcat(expt_name, "IFC");
  *perfect_performance = 100.0;
  *chance_performance = 100.0/(double)expt_opt;
}

/*******************************************************************************************/

void init_adaptive_params (int adaptive_opt, int* max_right_count, int* max_wrong_count, 
                           int* harder_direction, 
                           char* long_adaptive_name, char* short_adaptive_name)
{
/*
  *harder_direction =  1  if an increase in the stimulus value is supposedly harder in the task
  *harder_direction = -1  if a  decrease in the stimulus value is supposedly harder in the task

This can be generalised to m-up n-down (and vice versa, with both m > 1 and n > 1.
The function check_changepoint() can handle this.
NB that run lengths may increase  MAX_TRIALS_PER_RUN
*/
  strcpy(short_adaptive_name,"");
  strcpy( long_adaptive_name,"");

  if (adaptive_opt == ADAPTIVE_TWO_UP_ONE_DOWN) {
    strcpy(long_adaptive_name,"Adaptive two up one down");
    strcpy(short_adaptive_name,"2U1D");
    *max_right_count = 2;
    *max_wrong_count = 1;
    *harder_direction = 1;
  }
  if (adaptive_opt == ADAPTIVE_TWO_DOWN_ONE_UP) {
    strcpy(short_adaptive_name,"Adaptive two down one up");
    strcpy(short_adaptive_name,"2D1U");
    *max_right_count = 2;
    *max_wrong_count = 1;
    *harder_direction = -1;
  }
  if (adaptive_opt == ADAPTIVE_THREE_UP_ONE_DOWN) {
    strcpy(short_adaptive_name,"Adaptive three up one down");
    strcpy(short_adaptive_name,"3U1D");
    *max_right_count = 3;
    *max_wrong_count = 1;
    *harder_direction = 1;
  }
  if (adaptive_opt == ADAPTIVE_THREE_DOWN_ONE_UP) {
    strcpy(short_adaptive_name,"Adaptive three down one up");
    strcpy(short_adaptive_name,"3D1U");
    *max_right_count = 3;
    *max_wrong_count = 1;
    *harder_direction = -1;
  }
  if (adaptive_opt == ADAPTIVE_FOUR_UP_ONE_DOWN) {
    strcpy(short_adaptive_name,"Adaptive four up one down");
    strcpy(short_adaptive_name,"4U1D");
    *max_right_count = 4;
    *max_wrong_count = 1;
    *harder_direction = 1;
  }
  if (adaptive_opt == ADAPTIVE_FOUR_DOWN_ONE_UP) {
    strcpy(short_adaptive_name,"Adaptive four down one up");
    strcpy(short_adaptive_name,"4D1U");
    *max_right_count = 4;
    *max_wrong_count = 1;
    *harder_direction = -1;
  }

  if (adaptive_opt == ADAPTIVE_TEST) {
    *max_right_count = 2;
    *max_wrong_count = 1;
    *harder_direction = 1;
  }

}

/*******************************************************************************************/

void check_changepoint (run_arr_type curr_trialblock, 
                        reversal_arr_type reversal_arr,
                        int adaptive_opt, 
                        int geometric_opt, 
                        int curr_trial, 
                        int* right_count, 
                        int* wrong_count, 
                        int max_right_count, 
                        int max_wrong_count, 
                        int harder_direction, 
                        int* changepoint_idx, 
                        int* is_changepoint, 
                        int* hard_easy_same_opt, 
                        int* prev_direction, 
                        int prev_changepoint_idx, 
                        int* prev_level_change_idx, 
                        int* is_level_change)
{                         
  int i, decision, easier_direction;

  decision = curr_trialblock[curr_trial].right_wrong;
  easier_direction = -harder_direction;
  
  *hard_easy_same_opt = 0;  
/*  No change by default. It's only set to harder_direction or easier_direction below
    once the respective changepoint criterion has been met.
*/

  if (decision == CORRECT) {
    if (curr_trial >= *prev_level_change_idx + max_right_count) {
      for (i=0; i<max_right_count; i++) 
        if (curr_trialblock[curr_trial-i].right_wrong == CORRECT)
          (*right_count)++;
      if (*right_count == max_right_count) {
        *right_count = 0;
        *hard_easy_same_opt = harder_direction;    /* direction for next trial */
        *prev_level_change_idx = curr_trial;
        *is_level_change = 1;
        if (*prev_direction != harder_direction) {
          (*changepoint_idx)++;
          *is_changepoint = 1;
          reversal_arr[*changepoint_idx].level = curr_trialblock[curr_trial].level;
          reversal_arr[*changepoint_idx].trial_idx = curr_trial;
          reversal_arr[*changepoint_idx].direction = *prev_direction;
          *prev_direction = harder_direction;
        }
      }
    }
  } 
  else {  /* (decision == INCORRECT) */
    if (curr_trial >= *prev_level_change_idx + max_wrong_count) {
      for (i=0; i<max_wrong_count; i++) 
        if (curr_trialblock[curr_trial-i].right_wrong == INCORRECT)
          (*wrong_count)++;
      if (*wrong_count == max_wrong_count) {
        *wrong_count = 0;
        *hard_easy_same_opt = easier_direction;     /* direction for next trial */
        *prev_level_change_idx = curr_trial;
        *is_level_change = 1;
        if (*prev_direction != easier_direction) {
          (*changepoint_idx)++;
          *is_changepoint = 1;
          reversal_arr[*changepoint_idx].level = curr_trialblock[curr_trial].level;
          reversal_arr[*changepoint_idx].trial_idx = curr_trial;
          reversal_arr[*changepoint_idx].direction = *prev_direction;
          *prev_direction = easier_direction;
        }
      }
    }
  }
}

/*******************************************************************************************/

int run_block (pfn_ptr_type pfn_ptr, run_arr_type curr_trialblock, 
							 reversal_arr_type reversal_arr, 
               double starting_level, double starting_step_size, double final_step_size, 
               int max_num_reversals, int first_main_reversal_idx, 
               int expt_opt, int adaptive_opt, int initialise_opt, int display_opt, 
               FILE *fptr, 
               char* expt_name, char* long_adaptive_name, char* short_adaptive_name, 
               long int* local_first_seed, int run_idx_number, int max_num_repeats)

{
  int i, j;
  int decision, hard_easy_same_opt, prev_direction, harder_direction, max_right_count, max_wrong_count;

	int halting_criterion = FALSE;
	int geometric_opt = FALSE;
  int retval = -1;
  
  int curr_trial = -1;
	int reversal_idx	= -1;
  int changepoint_idx = -1;
  int prev_changepoint_idx = 0;
  int prev_level_change_idx = -1;

  int right_count = 0;
  int wrong_count = 0;
  int is_level_change = 0;
  int is_changepoint = 1;

	double curr_level, curr_step_size, temp, cutpoint;
  char comparison[8];
  static char change_char[4];

  curr_level = starting_level;
	curr_step_size = starting_step_size;
  
/*  perfect_performance and chance_performance are global variables, 
    set in value via init_expt_opt() and used in various pfns.
*/
  if (initialise_opt != 0) {
    init_expt_opt (expt_opt, &perfect_performance, &chance_performance, expt_name);
    init_adaptive_params (adaptive_opt, &max_right_count, &max_wrong_count, 
                          &harder_direction, long_adaptive_name, short_adaptive_name);
  }
/* 
Now pre-run the first trial to see if it will be CORRECT or INCORRECT, 
so that prev_direction is initialised to the direction of the first trial
and the very first trial will not count as a changepoint in check_changepoint().
*/
  if ((100.0*((double)random())/RAND_MAX) <= pfn_ptr(curr_level))
    prev_direction = harder_direction;
	else 
    prev_direction = -harder_direction;  /* i.e. easier_direction */
  srandom(*local_first_seed);   
/* resets the seed to what it was */

  if (display_opt == 1) {
    fprintf (fptr, "\nstarting run %d of %d\n", run_idx_number+1, max_num_repeats);
    fprintf (fptr, "start seed = %10ld,  RAND_MAX = %ld\n\n", *local_first_seed, RAND_MAX);
    fprintf (fptr, "%s  %s\n\n", expt_name, short_adaptive_name);
  	fprintf(fptr, "  * changepoint,   . change level,   ^ increasing,   v decreasing,    < > cutpoint_vs_temp\n\n"); 
  	fprintf(fptr, "trial   X correct change_idx cutpoint    temp   step_size  direction\n"); 
  }

  change_char[0] = ' ';
  change_char[1] = ' ';
  change_char[2] = ' ';

  while (halting_criterion == FALSE) {
    curr_trial++;
    if (curr_trial >= MAX_TRIALS_PER_RUN) {
      halting_criterion = TRUE;
      fprintf (fptr, "\n EARLY EXIT:  curr_trial = %d >= %d = MAX_TRIALS_PER_RUN\n", curr_trial, MAX_TRIALS_PER_RUN);
      retval = -1;
      break;
    }

		cutpoint = pfn_ptr(curr_level);
    temp = 100.0*((double)random())/RAND_MAX;  
	  if (temp <= cutpoint)
			decision = CORRECT;
		else 
			decision = INCORRECT;
		curr_trialblock[curr_trial].level = curr_level;
		curr_trialblock[curr_trial].right_wrong = decision;

    if (is_changepoint == 1) {
      right_count = 0;
      wrong_count = 0;
      is_changepoint = 0;
    }

    check_changepoint (curr_trialblock, 
                       reversal_arr, 
                       adaptive_opt, 
                       geometric_opt, 
                       curr_trial, 
                       &right_count, 
                       &wrong_count, 
                       max_right_count, 
                       max_wrong_count, 
                       harder_direction, 
                       &changepoint_idx, 
                       &is_changepoint, 
                       &hard_easy_same_opt, 
                       &prev_direction, 
                       prev_changepoint_idx, 
                       &prev_level_change_idx, 
                       &is_level_change);

    change_char[0] = ' ';
    change_char[1] = ' ';
    if (hard_easy_same_opt != 0)
      if (hard_easy_same_opt == 1)
        change_char[2] = '^';
      else
        change_char[2] = 'v';

    if (is_level_change == 1) {
      is_level_change = 0;
      prev_level_change_idx = curr_trial;
      change_char[1] = '.';
    }

    if (is_changepoint == 1) {
      change_char[0] = '*';
    }

	  if (temp < cutpoint)
      strcpy(comparison,"  >");
    else if (temp > cutpoint)
      strcpy(comparison,"<  ");
    else  
      strcpy(comparison," = ");

    if (display_opt == 1)
  		fprintf (fptr, "%3d   %3.0f    %d %c%c%c   %2d       %6.2f %s %5.2f     %2.0f         %2d\n", 
				curr_trial, 
				curr_trialblock[curr_trial].level,
				curr_trialblock[curr_trial].right_wrong, change_char[0], change_char[1], change_char[2], changepoint_idx, cutpoint, comparison, temp,  curr_step_size, 
        prev_direction);

	  if (changepoint_idx >= first_main_reversal_idx - 1)
      curr_step_size = final_step_size;

    if (hard_easy_same_opt == -1) 
      curr_level = curr_level - curr_step_size;
    else if (hard_easy_same_opt == 1) 
      curr_level = curr_level + curr_step_size;
    /*else curr_level unchanged*/

    if (changepoint_idx >= max_num_reversals - 1) {
      halting_criterion = TRUE;
      retval = 0;
      if (display_opt == 1)
        fprintf (fptr, "halting_criterion = TRUE\n");
    }
  }


  *local_first_seed = random();   /* update seed */
  return retval;
} 

/*******************************************************************************************/

main (int argc, char** argv)
{
  pfn_ptr_type  pfn_arr[MAX_NUM_PFNS];
  pfn_ptr_type  pfn_ptr;
  run_arr_type curr_trialblock;
  all_reversal_arr_type all_reversal_arr;
  all_run_arr_type  all_run_arr; 

  int i, j, sum;
	int run_okay;
  int pfn_opt;
  time_t t1,t2;
  int duration;

  int display_opt;
  int save_separately_flag;
  int write_header_flag_a;
  int write_header_flag_b;
  
  double pfn_coeffs[N_PFN_COEFFS];

  double starting_level;
  double starting_step_size; 
  double final_step_size;

  int max_num_conds;
  int max_num_repeats;
  int max_num_reversals;
  int reversal_highest_index;
  int num_reversals_used;
  int first_main_reversal_idx;
  int last_main_reversal_idx;

  int coeff_set_idx;
  double stdev_cutoff_value;
  long int start_seed, end_seed, seed_current_block;
  /* (int==long int) on linux i386 at the mo'. Be wary if porting. */

  int error_ct = 0;
  int exit_ct  = 0;
  int overrun_ct   = 0;
  int recompile_ct = 0;
  
/*  int adaptive_opt = ADAPTIVE_TWO_DOWN_ONE_UP;   */  
/*  int adaptive_opt = ADAPTIVE_FOUR_UP_ONE_DOWN;  */
/*  int adaptive_opt = ADAPTIVE_THREE_UP_ONE_DOWN; */
  int adaptive_opt = ADAPTIVE_TWO_UP_ONE_DOWN; 
  int expt_opt = EXPT_2IFC;

  FILE *fptr_a;
  FILE *fptr_b;
  char outfile_a[1000];
  char outfile_b[1000];
  char headerstr[1000];
  char pfn_coefficient_filename[1000];
  char  long_adaptive_name[40];
  char short_adaptive_name[40];
  char expt_name[40];
  char pfn_name[40];

  struct stat buf;
  int check_for_file;
  int fptr_a_isopen;
  int fptr_b_isopen;
  
  reversal_highest_index  = REVERSAL_HIGHEST_IDX;
  max_num_conds           = MAX_NUM_CONDS;
  max_num_repeats         = MAX_NUM_REPEATS;
  first_main_reversal_idx = FIRST_MAIN_REVERSAL_IDX;
  num_reversals_used      = NUM_REVERSALS_USED;
  max_num_reversals       = MAX_NUM_REVERSALS;
  last_main_reversal_idx  = LAST_MAIN_REVERSAL_IDX;

  starting_level     = INITIAL_LEVEL;
  starting_step_size = INITIAL_STEP_SIZE; 
  final_step_size    = FINAL_STEP_SIZE;

  pfn_opt            = PFN_OPT;
  stdev_cutoff_value = STDEV_CUTOFF_VALUE;
  display_opt        = DISPLAY_OPT;
  save_separately_flag = SAVE_SEPARATELY_FLAG;
  coeff_set_idx      = COEFF_SET_IDX;
  strcpy (pfn_coefficient_filename, "test_pfn_coeffs_pfn1.txt");

/* Usage output */
  if ( ((argc != 1) && (argc != 2) && (argc != N_PROG_PARAMS) && (argc != N_PROG_PARAMS+1)) || 
       ((argc == 2) && (argv[1][0] == '-') && (argv[1][1] == 'h')) || 
       ((argc == 2) && (argv[1][0] == '-') && (argv[1][1] == '-') && (argv[1][2] == 'h')) ) {  
    exit_ct++;
    fprintf(stderr, "\n\nUsage:\n\n%s\n", argv[0]);
    fprintf(stderr, " or\n");
    fprintf(stderr, "%s rand_seed    (integer-valued seed)\n", argv[0], argv[1]);
    fprintf(stderr, " or...\n");
    fprintf(stderr, "%s\n  n_initial_changepoints  n_final_changepoints  n_repeats  \n", argv[0]);
    fprintf(stderr, "  starting_level  starting_step_size  final_step_size\n");
    fprintf(stderr, "  expt_opt  adaptive_opt  pfn_opt\n");
    fprintf(stderr, "  stdev_cutoff_value\n");
    fprintf(stderr, "  save_separately_flag  display_opt\n");
    fprintf(stderr, "  seed\n");
    fprintf(stderr, "  coeff_set_idx  pfn_coefficient_filename\n");
    fprintf(stderr, "  [results_filename]\n(all on one line)\n\n");
    fprintf(stderr, "Short form:\n");
    fprintf(stderr, "  uses default parameters given in source code;\n  output is to stdout.\n");
    fprintf(stderr, "Long form:\n");
    fprintf(stderr, "  output is to stdout if results_filename is not specified.\n\n");
    fprintf(stderr, "pfn_coefficient_filename must exist and be in the correct format\n");
    fprintf(stderr, "for the type of psychometric function specified by pfn_opt.\n\n\n");
    exit (exit_ct);
  }

  if ( ((argc == 2) && (argv[1][0] == '-') && (argv[1][1] == 'v')) || 
       ((argc == 2) && (argv[1][0] == '-') && (argv[1][1] == '-') && (argv[1][2] == 'v')) ) {  
    exit_ct++;
    fprintf(stdout, "\npsychom_siml_guts  Version 1.00\n\n");
    exit (exit_ct);
  }


  strcpy (headerstr, 
"#stdev_cutoff\tN_runs\tout_of_possible_N_runs\tpercentage_of_runs\tstarting_level\tstarting_step_size\tfinal_step_size\t\
mean\tstderr\tstdev\t\
global_min\tglobal_max\trms_stdev\t\
initial_changepoints\tfinal_changepoints\t\
mean_start\tstdev_start\tmean_end\tstdev_end\tmean_length\tstdev_length\t\
min_array_end\tmax_array_end\t\
seed_at_start\tseed_at_end\t\
expt_opt\texpt_name\tadaptive_opt\tadaptive_name\tpfn_opt\tpfn_name\t\
pfn_coefficient_filename\tcoeff_set_idx\tpfn_coefficients...");
/*
NB: this headerstr must be consistent with the parameters written at the end of collate_reversal_arrs()
    The "#" is there to make the header a comment line in gnuplot
*/

  start_seed = 7777777;
  if (argc == 2) {
    start_seed = atol(argv[1]);    /* NB atol() not atoi() */
  }

  if (argc >= N_PROG_PARAMS) {
    first_main_reversal_idx = atoi(argv[ 1]);
    num_reversals_used      = atoi(argv[ 2]);
    max_num_repeats         = atoi(argv[ 3]);
    starting_level          = atof(argv[ 4]);
    starting_step_size      = atof(argv[ 5]);
    final_step_size         = atof(argv[ 6]);
    expt_opt                = atoi(argv[ 7]);
    adaptive_opt            = atoi(argv[ 8]);
    pfn_opt                 = atoi(argv[ 9]);
    stdev_cutoff_value      = atof(argv[10]);
    save_separately_flag    = atoi(argv[11]);
    display_opt             = atoi(argv[12]);
    start_seed              = atol(argv[13]);  /* NB atol() not atoi() */
    coeff_set_idx           = atoi(argv[14]);
    strcpy(pfn_coefficient_filename, argv[15]);
  }
 /* argv[0] is the program name "psychom_siml_guts" */
 /* argv[16] is the optional results_filename. It's treated differently later. */

  max_num_reversals = first_main_reversal_idx + num_reversals_used;
  last_main_reversal_idx = max_num_reversals - 1;

  error_ct = check_errors_in_main (max_num_conds, max_num_repeats, 
               max_num_reversals, num_reversals_used,
               first_main_reversal_idx, last_main_reversal_idx, reversal_highest_index, 
               argc, expt_opt, display_opt, save_separately_flag);
  if (error_ct != 0) {
    exit_ct++;
    fprintf(stderr, "\nParameter problems in main().  Early exit.\n\n");
    exit(exit_ct);
  }
  
/*  Leave these calls here. initialize_pfn() later below needs init_expt_opt() 
    to have been called first. */
  initialize_all_run_array      (all_run_arr,      max_num_conds, max_num_repeats);
  initialize_all_reversal_array (all_reversal_arr, max_num_conds, max_num_repeats);
  init_expt_opt (expt_opt, &perfect_performance, &chance_performance, expt_name);
  if (error_ct > 0) {
    exit_ct++;
    fprintf(stderr, "Argument error in main(). Early exit\n\n");
    clean_up (all_reversal_arr, all_run_arr, max_num_conds, max_num_repeats,
              fptr_a_isopen, fptr_b_isopen, fptr_a, fptr_b);
    exit(exit_ct);
  }

  error_ct = load_pfn_coeffs(pfn_opt, pfn_coeffs, pfn_coefficient_filename, coeff_set_idx);    
  if (error_ct != 0) {
    exit_ct++;
    fprintf(stderr, "\n\nProblems loading pfn coefficient file. Early exit\n");
    clean_up (all_reversal_arr, all_run_arr, max_num_conds, max_num_repeats,
              fptr_a_isopen, fptr_b_isopen, fptr_a, fptr_b);
    exit(exit_ct);
  }

  pfn_ptr = initialize_pfn (pfn_opt, pfn_arr, pfn_coeffs, 1, pfn_name);
  if (pfn_ptr == NULL) {
    exit_ct++;
    fprintf(stderr, "Problems in initialize_pfn().  Early exit.\n\n");
    clean_up (all_reversal_arr, all_run_arr, max_num_conds, max_num_repeats,
              fptr_a_isopen, fptr_b_isopen, fptr_a, fptr_b);
    exit(exit_ct);
  }

  fptr_a = stdout;
  fptr_b = stdout;
  strcpy (outfile_a, "");
  strcpy (outfile_b, "");
  if (argc == N_PROG_PARAMS+1) {
    strcpy (outfile_a, argv[N_PROG_PARAMS]);
    strcpy (outfile_b, argv[N_PROG_PARAMS]);
    if (save_separately_flag == 1) {
      strcat (outfile_a, ".stdev_unlimited.txt");
      strcat (outfile_b, ".stdev_is_finite.txt");
    } 
  }

/*
below: 
write header(s) by default, but if the results file(s) already exist, 
then do not write header(s).
This is geared towards display_opt = 5
*/
  error_ct = 0;
  write_header_flag_a = 0;
  if (save_separately_flag == 0)
    write_header_flag_b = 0;
  else
    write_header_flag_b = 1;

  if (strlen(outfile_a) == 0) {  /* writing to stdout */
    if (display_opt >= 4) {
      write_header_flag_a = 1;
      write_header_flag_b = 0;
    }
  } else {                      /* possibly writing to disk file */
    write_header_flag_a = 0;
    write_header_flag_b = 0;
    check_for_file = stat(outfile_a, &buf);
    if (check_for_file < 0)
      if (errno == ENOENT)
        write_header_flag_a = 1;
      else 
        error_ct++;
    if (save_separately_flag == 1) {
      check_for_file = stat(outfile_b, &buf);
      if (check_for_file < 0)
        if (errno == ENOENT)
          write_header_flag_b = 1;
        else 
          error_ct++;
    } else {
      write_header_flag_b = 0;  /*redundant, but what the hey */
      strcpy (outfile_b, "");
    }
  }

  if (error_ct > 0) {
    exit_ct++;
    fprintf(stderr, "\n\nResults-file error in main(). Early exit\n");
    clean_up (all_reversal_arr, all_run_arr, max_num_conds, max_num_repeats,
              fptr_a_isopen, fptr_b_isopen, fptr_a, fptr_b);
    exit(exit_ct);
  }

/*printf ("write_header_flag_a = %d, write_header_flag_b = %d\n", write_header_flag_a, write_header_flag_b);*/
    
  fptr_a_isopen = 0;
  fptr_b_isopen = 0;
  if (strlen(outfile_a) == 0) {
    if (write_header_flag_a == 1)    /* writing to stdout */
      fprintf (stdout, "%s\n", headerstr);
  } else {
    if (write_header_flag_a == 1) {
      fptr_a = fopen (outfile_a, "w");
      if ((display_opt == 4) || (display_opt == 5)) 
        fprintf (fptr_a, "%s\n", headerstr);
    } else {
      fptr_a = fopen (outfile_a, "a");
    }
    fptr_a_isopen = 1;
    if (save_separately_flag == 1) {
      if (write_header_flag_b == 1) {
        fptr_b = fopen (outfile_b, "w");
        if ((display_opt == 4) || (display_opt == 5)) 
          fprintf (fptr_b, "%s\n", headerstr);
      } else {
        fptr_b = fopen (outfile_b, "a");
      }
      fptr_b_isopen = 1;
    } else {
      fptr_b = fptr_a;
    }
  }

/*
  strcpy (outfile, argv[N_PROG_PARAMS]);
  check_for_file = stat(outfile, &buf);
  printf ("check_for_file = %d,  errno = %d\n", check_for_file, errno);
*/

  (void) time(&t1);

  srandom(start_seed);
  seed_current_block = start_seed;

  overrun_ct = 0;
  for (i=0; i<max_num_conds; i++) { /* condition numbers are not used at present, so i=0 only. */
    for (j=0; j<max_num_repeats; j++) {
      init_run_array (&(all_run_arr[i][j]), MAX_TRIALS_PER_RUN);
      init_reversal_array (&(all_reversal_arr[i][j]), REVERSAL_HIGHEST_IDX+1);
      run_okay = run_block (pfn_ptr, all_run_arr[i][j], all_reversal_arr[i][j], 
                            starting_level, starting_step_size, final_step_size, 
                            max_num_reversals, first_main_reversal_idx, 
                            expt_opt, adaptive_opt, 1, display_opt, 
                            fptr_a, 
                            expt_name, long_adaptive_name, short_adaptive_name, 
                            &seed_current_block, j, max_num_repeats);

      if (run_okay == 0) {
        crunch_reversal_arr (all_reversal_arr[i][j], first_main_reversal_idx, last_main_reversal_idx);
        if ((1 <= display_opt) && (display_opt <= 2)) 
          display_reversal_arr (all_reversal_arr[i][j], reversal_highest_index, max_num_reversals, 
                                i, j, max_num_repeats,
                                display_opt, fptr_a, seed_current_block);
      } else {
        overrun_ct++;
        if (overrun_ct <= MAX_NUM_OVERRUNS) 
          printf("\nRun length exceeds array bounds:  overrun_ct = %d.  Trying again...\n", overrun_ct);
        else  
          printf("\nRun length exceeds array bounds:  overrun_ct = %d.  Giving up.\n", overrun_ct);
        j--;
      }
      if (overrun_ct > MAX_NUM_OVERRUNS) {  /* exit both the i and j loops */
        i = j = INT_MAX-2; /* to break out of both loops. Also see note below. */
        overrun_ct = -1;
      }
    }
  }

/* loop-breaking curio: don't set i = j = INT_MAX because i++ in the for loop 
 would then wrap i around from INT_MAX to INT_MIN (e.g. +2147483647 to -2147483648)
 which leaves i<0 true, and hence you never break out of the loop.  */

  
  if (overrun_ct < 0) {
    exit_ct++;
    fprintf(stderr, "Run array length exceeded too many times. Early exit\n\n");
    clean_up (all_reversal_arr, all_run_arr, max_num_conds, max_num_repeats,
              fptr_a_isopen, fptr_b_isopen, fptr_a, fptr_b);
    exit(exit_ct);
  } else {
    end_seed = seed_current_block;
    if ((display_opt == 5) && (argc == N_PROG_PARAMS+1)) {
      collate_reversal_arrs (all_reversal_arr, max_num_conds, max_num_repeats, 
                             first_main_reversal_idx, last_main_reversal_idx, num_reversals_used, 
                             stdev_cutoff_value, 
                             starting_level, starting_step_size, final_step_size, 
                             start_seed, end_seed, 
                             expt_opt, adaptive_opt, pfn_opt, 3,
                             expt_name, short_adaptive_name, pfn_name,
                             stdout, stdout,
                             pfn_coefficient_filename, coeff_set_idx, pfn_coeffs);
      collate_reversal_arrs (all_reversal_arr, max_num_conds, max_num_repeats, 
                             first_main_reversal_idx, last_main_reversal_idx, num_reversals_used, 
                             stdev_cutoff_value, 
                             starting_level, starting_step_size, final_step_size, 
                             start_seed, end_seed, 
                             expt_opt, adaptive_opt, pfn_opt, 4,
                             expt_name, short_adaptive_name, pfn_name,
                             fptr_a, fptr_b,
                             pfn_coefficient_filename, coeff_set_idx, pfn_coeffs);
    } else {
      collate_reversal_arrs (all_reversal_arr, max_num_conds, max_num_repeats, 
                             first_main_reversal_idx, last_main_reversal_idx, num_reversals_used, 
                             stdev_cutoff_value, 
                             starting_level, starting_step_size, final_step_size, 
                             start_seed, end_seed, 
                             expt_opt, adaptive_opt, pfn_opt, display_opt, 
                             expt_name, short_adaptive_name, pfn_name,
                             fptr_a, fptr_b,
                             pfn_coefficient_filename, coeff_set_idx, pfn_coeffs);
    }  


    clean_up (all_reversal_arr, all_run_arr, max_num_conds, max_num_repeats,
              fptr_a_isopen, fptr_b_isopen, fptr_a, fptr_b);

    strcpy (outfile_a, "./psychom_randseed");
    strcpy (outfile_b, "./psychom_randseed_bak");
    check_for_file = rename (outfile_a, outfile_b);
  /*  printf ("rename(),  status = %d,  errno = %d\n", check_for_file, errno);*/
    fptr_a = fopen  (outfile_a, "w");
    fprintf (fptr_a, "Globally across runs:  start_seed\tend_seed\n");
    fprintf (fptr_a, "%ld\t%ld", start_seed, end_seed);
    fclose (fptr_a);
  }

  (void) time(&t2);
  duration = difftime(t2, t1);

  printf("\nTime taken = %d seconds\n", duration);  
  printf("All done\n");
  exit(0);
}

/*******************************************************************************************/


